import importlib.metadata

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
]

templates_path = []

source_suffix = '.rst'
master_doc = 'index'

project = 'aetcd3'
copyright = '2020, Andrey Martyanov'
author = 'Andrey Martyanov'

version = release = importlib.metadata.version('aetcd3')

autodoc_member_order = 'bysource'

html_theme = 'sphinx_rtd_theme'
html_theme_options = {}
html_static_path = []



import asyncio
import functools
import inspect
import tempfile
import warnings

import aiofiles
import grpclib
import grpclib.client

from . import exceptions
from . import leases
from . import locks
from . import members
from . import rpc
from . import transactions
from . import utils
from . import watch


_EXCEPTIONS_BY_CODE = {
    grpclib.Status.INTERNAL: exceptions.InternalServerError,
    grpclib.Status.UNAVAILABLE: exceptions.ConnectionFailedError,
    grpclib.Status.DEADLINE_EXCEEDED: exceptions.ConnectionTimeoutError,
    grpclib.Status.FAILED_PRECONDITION: exceptions.PreconditionFailedError,
}


def _translate_exception(error: grpclib.GRPCError):
    exc = _EXCEPTIONS_BY_CODE.get(error.status)
    if exc is not None:
        raise exc
    raise


class Transactions(object):
    def __init__(self):
        self.value = transactions.Value
        self.version = transactions.Version
        self.create = transactions.Create
        self.mod = transactions.Mod

        self.put = transactions.Put
        self.get = transactions.Get
        self.delete = transactions.Delete
        self.txn = transactions.Txn


class KVMetadata(object):
    def __init__(self, keyvalue, header):
        self.key = keyvalue.key
        self.create_revision = keyvalue.create_revision
        self.mod_revision = keyvalue.mod_revision
        self.version = keyvalue.version
        self.lease_id = keyvalue.lease
        self.response_header = header


class Status(object):
    def __init__(self, version, db_size, leader, raft_index, raft_term):
        self.version = version
        self.db_size = db_size
        self.leader = leader
        self.raft_index = raft_index
        self.raft_term = raft_term


class Alarm(object):
    """A cluster member alarm.

    :param alarm_type: Type of the alarm
    :param member_id: Cluster member ID
    """

    def __init__(self, alarm_type, member_id):
        self.alarm_type = alarm_type
        self.member_id = member_id


def _handle_errors(f):  # noqa: C901
    if inspect.isasyncgenfunction(f):
        async def handler(*args, **kwargs):
            try:
                async for data in f(*args, **kwargs):
                    yield data
            except grpclib.GRPCError as exc:
                _translate_exception(exc)
    elif inspect.iscoroutinefunction(f):
        async def handler(*args, **kwargs):
            try:
                return await f(*args, **kwargs)
            except grpclib.GRPCError as exc:
                _translate_exception(exc)
    else:
        def handler(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except grpclib.GRPCError as exc:
                _translate_exception(exc)

    return functools.wraps(f)(handler)


def _ensure_channel(f):
    if inspect.isasyncgenfunction(f):
        async def handler(*args, **kwargs):
            await args[0].open()
            async for data in f(*args, **kwargs):
                yield data
    elif inspect.iscoroutinefunction(f):
        async def handler(*args, **kwargs):
            await args[0].open()
            return await f(*args, **kwargs)
    else:
        raise TypeError

    return functools.wraps(f)(handler)


class Etcd3Client:
    """Client."""

    def __init__(
            self,
            host='localhost',
            port=2379,
            ca_cert=None,
            cert_key=None,
            cert_cert=None,
            timeout=None,
            user=None,
            password=None,
            grpc_options=None,
    ):
        self.host = host
        self.port = port
        self.timeout = timeout
        self.call_credentials = None
        self.transactions = Transactions()

        if grpc_options:
            warnings.warn("grpc_options can't be used with asyncio backend")

        cert_params = (cert_cert, cert_key)
        if any(cert_params) and None in cert_params:
            raise ValueError(
                'to use a secure channel ca_cert is required by itself, '
                'or cert_cert and cert_key must both be specified.')

        cred_params = (user, password)
        if any(cred_params) and None in cred_params:
            raise Exception(
                'if using authentication credentials both user and password '
                'must be specified.')

        self.ca_cert = ca_cert
        self.cert_key = cert_key
        self.cert_cert = cert_cert
        self.user = user
        self.password = password

        self._init_channel_attrs()

    def _init_channel_attrs(self):
        # These attributes will be assigned during opening of GRPC channel
        self.channel = None
        self.metadata = None
        self.uses_secure_channel = None
        self.auth_stub = None
        self.kvstub = None
        self.watcher = None
        self.clusterstub = None
        self.leasestub = None
        self.maintenancestub = None

    async def open(self):
        """Open GRPC channel."""
        if self.channel:
            return

        cert_params = [c is not None for c in (self.cert_cert, self.cert_key)]
        if self.ca_cert is not None:
            self.channel = grpclib.client.Channel(host=self.host, port=self.port, ssl=True)

            if all(cert_params):
                ca_bundle = tempfile.mktemp()
                async with aiofiles.open(ca_bundle, 'w') as cert_bundle:
                    for cf_path in (self.cert_cert, self.ca_cert):
                        async with aiofiles.open(cf_path) as cf:
                            await cert_bundle.write(await cf.read())
                    await cert_bundle.flush()
                    self.channel._ssl.load_cert_chain(ca_bundle, keyfile=self.cert_key)
            else:
                self.channel._ssl.load_cert_chain(self.ca_cert, keyfile=self.cert_key)

            self.uses_secure_channel = True
        else:
            self.uses_secure_channel = False
            self.channel = grpclib.client.Channel(host=self.host, port=self.port)

        cred_params = [c is not None for c in (self.user, self.password)]
        if all(cred_params):
            self.auth_stub = rpc.AuthStub(self.channel)
            auth_request = rpc.AuthenticateRequest(
                name=self.user,
                password=self.password,
            )

            resp = await self.auth_stub.Authenticate(auth_request, timeout=self.timeout)
            self.metadata = (('token', resp.token),)

        self.kvstub = rpc.KVStub(self.channel)
        self.watcher = watch.Watcher(rpc.WatchStub(self.channel), timeout=self.timeout)
        self.clusterstub = rpc.ClusterStub(self.channel)
        self.leasestub = rpc.LeaseStub(self.channel)
        self.maintenancestub = rpc.MaintenanceStub(self.channel)

    async def close(self):
        """Call the GRPC channel close semantics."""
        if self.channel:
            self.watcher.close()
            self.channel.close()
            self._init_channel_attrs()

    async def __aenter__(self):
        await self.open()
        return self

    async def __aexit__(self, *args):
        await self.close()

    @_handle_errors
    @_ensure_channel
    async def get(self, key, serializable=False):
        """Get the value of a key from etcd.

        example usage:

        .. code-block:: python

            import aetcd3
            etcd = aetcd3.client()
            await etcd.get('/thing/key')

        :param key: key in etcd to get
        :param serializable: whether to allow serializable reads. This can
            result in stale reads
        :returns: value of key and metadata
        :rtype: bytes, ``KVMetadata``
        """
        range_request = self._build_get_range_request(
            key,
            serializable=serializable)
        range_response = await self.kvstub.Range(
            range_request,
            timeout=self.timeout,
            metadata=self.metadata,
        )

        if range_response.count < 1:
            return None, None
        else:
            kv = range_response.kvs.pop()
            return kv.value, KVMetadata(kv, range_response.header)

    @_handle_errors
    @_ensure_channel
    async def get_prefix(self, key_prefix, sort_order=None, sort_target='key',
                         keys_only=False):
        """Get a range of keys with a prefix.

        :param key_prefix: first key in range

        :returns: sequence of (value, metadata) tuples
        """
        range_request = self._build_get_range_request(
            key=key_prefix,
            range_end=utils.prefix_range_end(utils.to_bytes(key_prefix)),
            sort_order=sort_order,
            sort_target=sort_target,
            keys_only=keys_only,
        )

        range_response = await self.kvstub.Range(
            range_request,
            timeout=self.timeout,
            metadata=self.metadata,
        )

        if range_response.count < 1:
            return
        else:
            for kv in range_response.kvs:
                yield (kv.value, KVMetadata(kv, range_response.header))

    @_handle_errors
    @_ensure_channel
    async def get_range(self, range_start, range_end, sort_order=None,
                        sort_target='key', **kwargs):
        """Get a range of keys.

        :param range_start: first key in range
        :param range_end: last key in range
        :returns: sequence of (value, metadata) tuples
        """
        range_request = self._build_get_range_request(
            key=range_start,
            range_end=range_end,
            sort_order=sort_order,
            sort_target=sort_target,
            **kwargs,
        )

        range_response = await self.kvstub.Range(
            range_request,
            timeout=self.timeout,
            metadata=self.metadata,
        )

        if range_response.count < 1:
            return
        else:
            for kv in range_response.kvs:
                yield (kv.value, KVMetadata(kv, range_response.header))

    @_handle_errors
    @_ensure_channel
    async def get_all(self, sort_order=None, sort_target='key',
                      keys_only=False):
        """Get all keys currently stored in etcd.

        :returns: sequence of (value, metadata) tuples
        """
        range_request = self._build_get_range_request(
            key=b'\0',
            range_end=b'\0',
            sort_order=sort_order,
            sort_target=sort_target,
            keys_only=keys_only,
        )

        range_response = await self.kvstub.Range(
            range_request,
            timeout=self.timeout,
            metadata=self.metadata,
        )

        if range_response.count < 1:
            return
        else:
            for kv in range_response.kvs:
                yield (kv.value, KVMetadata(kv, range_response.header))

    @_handle_errors
    @_ensure_channel
    async def put(self, key, value, lease=None, prev_kv=False):
        """Save a value to etcd.

        Example usage:

        .. code-block:: python

            import aetcd3
            etcd = aetcd3.client()
            await etcd.put('/thing/key', 'hello world')

        :param key: key in etcd to set
        :param value: value to set key to
        :type value: bytes
        :param lease: Lease to associate with this key.
        :type lease: either :class:`.Lease`, or int (ID of lease)
        :param prev_kv: return the previous key-value pair
        :type prev_kv: bool
        :returns: a response containing a header and the prev_kv
        :rtype: :class:`.rpc_pb2.PutResponse`
        """
        put_request = self._build_put_request(key, value, lease=lease,
                                              prev_kv=prev_kv)
        return await self.kvstub.Put(
            put_request,
            timeout=self.timeout,
            metadata=self.metadata,
        )

    @_handle_errors
    @_ensure_channel
    async def replace(self, key, initial_value, new_value):
        """Atomically replace the value of a key with a new value.

        This compares the current value of a key, then replaces it with a new
        value if it is equal to a specified value. This operation takes place
        in a transaction.

        :param key: key in etcd to replace
        :param initial_value: old value to replace
        :type initial_value: bytes
        :param new_value: new value of the key
        :type new_value: bytes
        :returns: status of transaction, ``True`` if the replace was
                  successful, ``False`` otherwise
        :rtype: bool
        """
        status, _ = await self.transaction(
            compare=[
                self.transactions.value(key) == initial_value,
            ],
            success=[
                self.transactions.put(key, new_value),
            ],
            failure=[
            ],
        )

        return status

    @_handle_errors
    @_ensure_channel
    async def delete(self, key, prev_kv=False, return_response=False):
        """Delete a single key in etcd.

        :param key: key in etcd to delete
        :param prev_kv: return the deleted key-value pair
        :type prev_kv: bool
        :param return_response: return the full response
        :type return_response: bool
        :returns: True if the key has been deleted when
                  ``return_response`` is False and a response containing
                  a header, the number of deleted keys and prev_kvs when
                  ``return_response`` is True
        """
        delete_request = self._build_delete_request(key, prev_kv=prev_kv)
        delete_response = await self.kvstub.DeleteRange(
            delete_request,
            timeout=self.timeout,
            metadata=self.metadata,
        )
        if return_response:
            return delete_response
        return delete_response.deleted >= 1

    @_handle_errors
    @_ensure_channel
    async def delete_prefix(self, prefix):
        """Delete a range of keys with a prefix in etcd."""
        delete_request = self._build_delete_request(
            prefix,
            range_end=utils.prefix_range_end(utils.to_bytes(prefix)),
        )
        return await self.kvstub.DeleteRange(
            delete_request,
            timeout=self.timeout,
            metadata=self.metadata,
        )

    @_handle_errors
    @_ensure_channel
    async def status(self):
        """Get the status of the responding member."""
        status_request = rpc.StatusRequest()
        status_response = await self.maintenancestub.Status(
            status_request,
            timeout=self.timeout,
            metadata=self.metadata,
        )

        async for m in self.members():
            if m.id == status_response.leader:
                leader = m
                break
        else:
            # raise exception?
            leader = None

        return Status(status_response.version,
                      status_response.dbSize,
                      leader,
                      status_response.raftIndex,
                      status_response.raftTerm)

    @_handle_errors
    @_ensure_channel
    async def add_watch_callback(self, *args, **kwargs):
        """Watch a key or range of keys and call a callback on every event.

        If timeout was declared during the client initialization and
        the watch cannot be created during that time the method raises
        a ``WatchTimedOut`` exception.

        :param key: key to watch
        :param callback: callback function

        :returns: watch_id. Later it could be used for cancelling watch.
        """
        try:
            return await self.watcher.add_callback(*args, **kwargs)
        except asyncio.QueueEmpty:
            raise exceptions.WatchTimedOut()

    @_handle_errors
    @_ensure_channel
    async def watch(self, key, **kwargs):
        """Watch a key.

        Example usage:

        .. code-block:: python

            events_iterator, cancel = await etcd.watch('/doot/key')
            async for event in events_iterator:
                print(event)

        :param key: key to watch

        :returns: tuple of ``events_iterator`` and ``cancel``.
                  Use ``events_iterator`` to get the events of key changes
                  and ``cancel`` to cancel the watch request
        """
        event_queue = asyncio.Queue()
        watch_id = await self.add_watch_callback(
            key, event_queue.put,
            **kwargs,
        )
        canceled = asyncio.Event()

        async def cancel():
            canceled.set()
            await event_queue.put(None)
            await self.cancel_watch(watch_id)

        @_handle_errors
        async def iterator():
            while not canceled.is_set():
                event = await event_queue.get()
                if event is None:
                    canceled.set()
                if isinstance(event, Exception):
                    canceled.set()
                    raise event
                if not canceled.is_set():
                    yield event

        return iterator(), cancel

    @_handle_errors
    @_ensure_channel
    async def watch_prefix(self, key_prefix, **kwargs):
        """Watches a range of keys with a prefix."""
        kwargs['range_end'] = \
            utils.prefix_range_end(utils.to_bytes(key_prefix))
        return await self.watch(key_prefix, **kwargs)

    @_handle_errors
    @_ensure_channel
    async def watch_once(self, key, timeout=None, **kwargs):
        """Watch a key and stops after the first event.

        If the timeout was specified and event didn't arrived method
        will raise ``WatchTimedOut`` exception.

        :param key: key to watch
        :param timeout: (optional) timeout in seconds.
        :returns: :class:`aetcd3.Event`
        """
        event_queue = asyncio.Queue()

        watch_id = await self.add_watch_callback(key, event_queue.put,
                                                 **kwargs)

        try:
            return await asyncio.wait_for(event_queue.get(), timeout)
        except (asyncio.QueueEmpty, asyncio.TimeoutError):
            raise exceptions.WatchTimedOut()
        finally:
            await self.cancel_watch(watch_id)

    @_handle_errors
    @_ensure_channel
    async def watch_prefix_once(self, key_prefix, timeout=None, **kwargs):
        """Watches a range of keys with a prefix and stops after the first event.

        If the timeout was specified and event didn't arrived method
        will raise ``WatchTimedOut`` exception.
        """
        kwargs['range_end'] = utils.prefix_range_end(utils.to_bytes(key_prefix))
        return await self.watch_once(key_prefix, timeout=timeout, **kwargs)

    @_handle_errors
    @_ensure_channel
    async def cancel_watch(self, watch_id):
        """Stop watching a key or range of keys.

        :param watch_id: watch_id returned by ``add_watch_callback`` method
        """
        await self.watcher.cancel(watch_id)

    @_handle_errors
    @_ensure_channel
    async def transaction(self, compare, success=None, failure=None):
        """Perform a transaction.

        Example usage:

        .. code-block:: python

            await etcd.transaction(
                compare=[
                    etcd.transactions.value('/doot/testing') == 'doot',
                    etcd.transactions.version('/doot/testing') > 0,
                ],
                success=[
                    etcd.transactions.put('/doot/testing', 'success'),
                ],
                failure=[
                    etcd.transactions.put('/doot/testing', 'failure'),
                ]
            )

        :param compare: A list of comparisons to make
        :param success: A list of operations to perform if all the comparisons
                        are true
        :param failure: A list of operations to perform if any of the
                        comparisons are false
        :return: A tuple of (operation status, responses)
        """
        compare = [c.build_message() for c in compare]

        success_ops = self._ops_to_requests(success)
        failure_ops = self._ops_to_requests(failure)

        transaction_request = rpc.TxnRequest(
            compare=compare,
            success=success_ops,
            failure=failure_ops,
        )
        txn_response = await self.kvstub.Txn(
            transaction_request,
            timeout=self.timeout,
            metadata=self.metadata,
        )

        responses = []
        for response in txn_response.responses:
            response_type = response.WhichOneof('response')
            if response_type in ['response_put', 'response_delete_range',
                                 'response_txn']:
                responses.append(response)

            elif response_type == 'response_range':
                range_kvs = []
                for kv in response.response_range.kvs:
                    range_kvs.append((kv.value,
                                      KVMetadata(kv, txn_response.header)))

                responses.append(range_kvs)

        return txn_response.succeeded, responses

    @_handle_errors
    @_ensure_channel
    async def lease(self, ttl, lease_id=None):
        """Create a new lease.

        All keys attached to this lease will be expired and deleted if the
        lease expires. A lease can be sent keep alive messages to refresh the
        ttl.

        :param ttl: Requested time to live
        :param lease_id: Requested ID for the lease

        :returns: new lease
        :rtype: :class:`.Lease`
        """
        lease_grant_request = rpc.LeaseGrantRequest(TTL=ttl, ID=lease_id)
        lease_grant_response = await self.leasestub.LeaseGrant(
            lease_grant_request,
            timeout=self.timeout,
            metadata=self.metadata,
        )
        return leases.Lease(lease_id=lease_grant_response.ID,
                            ttl=lease_grant_response.TTL,
                            etcd_client=self)

    @_handle_errors
    @_ensure_channel
    async def revoke_lease(self, lease_id):
        """Revoke a lease.

        :param lease_id: ID of the lease to revoke.
        """
        lease_revoke_request = rpc.LeaseRevokeRequest(ID=lease_id)
        await self.leasestub.LeaseRevoke(
            lease_revoke_request,
            timeout=self.timeout,
            metadata=self.metadata,
        )

    @_handle_errors
    @_ensure_channel
    async def refresh_lease(self, lease_id):
        return await self.leasestub.LeaseKeepAlive(
            [rpc.LeaseKeepAliveRequest(ID=lease_id)],
            timeout=self.timeout,
            metadata=self.metadata)

    @_handle_errors
    @_ensure_channel
    async def get_lease_info(self, lease_id, *, keys=True):
        # only available in etcd v3.1.0 and later
        ttl_request = rpc.LeaseTimeToLiveRequest(
            ID=lease_id,
            keys=keys,
        )
        return await self.leasestub.LeaseTimeToLive(
            ttl_request,
            timeout=self.timeout,
            metadata=self.metadata,
        )

    @_handle_errors
    def lock(self, name, ttl=60):
        """Create a new lock.

        :param name: name of the lock
        :type name: string or bytes
        :param ttl: length of time for the lock to live for in seconds. The
                    lock will be released after this time elapses, unless
                    refreshed
        :type ttl: int
        :returns: new lock
        :rtype: :class:`.Lock`
        """
        return locks.Lock(name, ttl=ttl, etcd_client=self)

    @_handle_errors
    @_ensure_channel
    async def add_member(self, urls):
        """Add a member into the cluster.

        :returns: new member
        :rtype: :class:`members.Member`
        """
        member_add_request = rpc.MemberAddRequest(peerURLs=urls)

        member_add_response = await self.clusterstub.MemberAdd(
            member_add_request,
            timeout=self.timeout,
            metadata=self.metadata,
        )

        member = member_add_response.member
        return members.Member(
            member.ID,
            member.name,
            member.peerURLs,
            member.clientURLs,
            etcd_client=self,
        )

    @_handle_errors
    @_ensure_channel
    async def remove_member(self, member_id):
        """Remove an existing member from the cluster.

        :param member_id: ID of the member to remove
        """
        member_rm_request = rpc.MemberRemoveRequest(ID=member_id)
        await self.clusterstub.MemberRemove(
            member_rm_request,
            timeout=self.timeout,
            metadata=self.metadata,
        )

    @_handle_errors
    @_ensure_channel
    async def update_member(self, member_id, peer_urls):
        """Update the configuration of an existing member in the cluster.

        :param member_id: ID of the member to update
        :param peer_urls: new list of peer urls the member will use to
                          communicate with the cluster
        """
        member_update_request = rpc.MemberUpdateRequest(
            ID=member_id,
            peerURLs=peer_urls,
        )
        await self.clusterstub.MemberUpdate(
            member_update_request,
            timeout=self.timeout,
            metadata=self.metadata,
        )

    @_ensure_channel
    async def members(self):
        """List of all members associated with the cluster.

        :type: sequence of :class:`members.Member`

        """
        member_list_request = rpc.MemberListRequest()
        member_list_response = await self.clusterstub.MemberList(
            member_list_request,
            timeout=self.timeout,
            metadata=self.metadata,
        )

        for member in member_list_response.members:
            yield members.Member(
                member.ID,
                member.name,
                member.peerURLs,
                member.clientURLs,
                etcd_client=self,
            )

    @_handle_errors
    @_ensure_channel
    async def compact(self, revision, physical=False):
        """Compact the event history in etcd up to a given revision.

        All superseded keys with a revision less than the compaction revision
        will be removed.

        :param revision: revision for the compaction operation
        :param physical: if set to True, the request will wait until the
                         compaction is physically applied to the local database
                         such that compacted entries are totally removed from
                         the backend database
        """
        compact_request = rpc.CompactionRequest(
            revision=revision,
            physical=physical,
        )
        await self.kvstub.Compact(
            compact_request,
            timeout=self.timeout,
            metadata=self.metadata,
        )

    @_handle_errors
    @_ensure_channel
    async def defragment(self):
        """Defragment a member's backend database to recover storage space."""
        defrag_request = rpc.DefragmentRequest()
        await self.maintenancestub.Defragment(
            defrag_request,
            timeout=self.timeout,
            metadata=self.metadata,
        )

    @_handle_errors
    @_ensure_channel
    async def hash(self):
        """Return the hash of the local KV state.

        :returns: kv state hash
        :rtype: int
        """
        hash_request = rpc.HashRequest()
        return (await self.maintenancestub.Hash(hash_request)).hash

    @_handle_errors
    @_ensure_channel
    async def create_alarm(self, member_id=0):
        """Create an alarm.

        If no member id is given, the alarm is activated for all the
        members of the cluster. Only the `no space` alarm can be raised.

        :param member_id: The cluster member id to create an alarm to.
                          If 0, the alarm is created for all the members
                          of the cluster.
        :returns: list of :class:`.Alarm`
        """
        alarm_request = self._build_alarm_request('activate',
                                                  member_id,
                                                  'no space')
        alarm_response = await self.maintenancestub.Alarm(
            alarm_request,
            timeout=self.timeout,
            metadata=self.metadata,
        )

        return [Alarm(alarm.alarm, alarm.memberID)
                for alarm in alarm_response.alarms]

    @_handle_errors
    @_ensure_channel
    async def list_alarms(self, member_id=0, alarm_type='none'):
        """List the activated alarms.

        :param member_id:
        :param alarm_type: The cluster member id to create an alarm to.
                           If 0, the alarm is created for all the members
                           of the cluster.
        :returns: sequence of :class:`.Alarm`
        """
        alarm_request = self._build_alarm_request('get',
                                                  member_id,
                                                  alarm_type)
        alarm_response = await self.maintenancestub.Alarm(
            alarm_request,
            timeout=self.timeout,
            metadata=self.metadata,
        )

        for alarm in alarm_response.alarms:
            yield Alarm(alarm.alarm, alarm.memberID)

    @_handle_errors
    @_ensure_channel
    async def disarm_alarm(self, member_id=0):
        """Cancel an alarm.

        :param member_id: The cluster member id to cancel an alarm.
                          If 0, the alarm is canceled for all the members
                          of the cluster.
        :returns: List of :class:`.Alarm`
        """
        alarm_request = self._build_alarm_request('deactivate',
                                                  member_id,
                                                  'no space')
        alarm_response = await self.maintenancestub.Alarm(
            alarm_request,
            timeout=self.timeout,
            metadata=self.metadata,
        )

        return [Alarm(alarm.alarm, alarm.memberID)
                for alarm in alarm_response.alarms]

    @_handle_errors
    @_ensure_channel
    async def snapshot(self, file_obj):
        """Take a snapshot of the database.

        :param file_obj: A file-like object to write the database contents in.
        """
        snapshot_request = rpc.SnapshotRequest()
        snapshot_responses = await self.maintenancestub.Snapshot(
            snapshot_request,
            timeout=self.timeout,
            metadata=self.metadata,
        )

        for response in snapshot_responses:
            file_obj.write(response.blob)

    @staticmethod
    def _build_get_range_request(key,
                                 range_end=None,
                                 limit=None,
                                 revision=None,
                                 sort_order=None,
                                 sort_target='key',
                                 serializable=False,
                                 keys_only=False,
                                 count_only=None,
                                 min_mod_revision=None,
                                 max_mod_revision=None,
                                 min_create_revision=None,
                                 max_create_revision=None):
        range_request = rpc.RangeRequest()
        range_request.key = utils.to_bytes(key)
        range_request.keys_only = keys_only
        if range_end is not None:
            range_request.range_end = utils.to_bytes(range_end)

        if sort_order is None:
            range_request.sort_order = rpc.RangeRequest.NONE
        elif sort_order == 'ascend':
            range_request.sort_order = rpc.RangeRequest.ASCEND
        elif sort_order == 'descend':
            range_request.sort_order = rpc.RangeRequest.DESCEND
        else:
            raise ValueError('unknown sort order: "{}"'.format(sort_order))

        if sort_target is None or sort_target == 'key':
            range_request.sort_target = rpc.RangeRequest.KEY
        elif sort_target == 'version':
            range_request.sort_target = rpc.RangeRequest.VERSION
        elif sort_target == 'create':
            range_request.sort_target = rpc.RangeRequest.CREATE
        elif sort_target == 'mod':
            range_request.sort_target = rpc.RangeRequest.MOD
        elif sort_target == 'value':
            range_request.sort_target = rpc.RangeRequest.VALUE
        else:
            raise ValueError('sort_target must be one of "key", '
                             '"version", "create", "mod" or "value"')

        range_request.serializable = serializable

        return range_request

    @staticmethod
    def _build_put_request(key, value, lease=None, prev_kv=False):
        put_request = rpc.PutRequest()
        put_request.key = utils.to_bytes(key)
        put_request.value = utils.to_bytes(value)
        put_request.lease = utils.lease_to_id(lease)
        put_request.prev_kv = prev_kv

        return put_request

    @staticmethod
    def _build_delete_request(key,
                              range_end=None,
                              prev_kv=False):
        delete_request = rpc.DeleteRangeRequest()
        delete_request.key = utils.to_bytes(key)
        delete_request.prev_kv = prev_kv

        if range_end is not None:
            delete_request.range_end = utils.to_bytes(range_end)

        return delete_request

    def _ops_to_requests(self, ops):
        """
        Return a list of grpc requests.

        Returns list from an input list of etcd3.transactions.{Put, Get,
        Delete, Txn} objects.
        """
        request_ops = []
        for op in ops:
            if isinstance(op, transactions.Put):
                request = self._build_put_request(op.key, op.value,
                                                  op.lease, op.prev_kv)
                request_op = rpc.RequestOp(request_put=request)
                request_ops.append(request_op)

            elif isinstance(op, transactions.Get):
                request = self._build_get_range_request(op.key, op.range_end)
                request_op = rpc.RequestOp(request_range=request)
                request_ops.append(request_op)

            elif isinstance(op, transactions.Delete):
                request = self._build_delete_request(op.key, op.range_end,
                                                     op.prev_kv)
                request_op = rpc.RequestOp(request_delete_range=request)
                request_ops.append(request_op)

            elif isinstance(op, transactions.Txn):
                compare = [c.build_message() for c in op.compare]
                success_ops = self._ops_to_requests(op.success)
                failure_ops = self._ops_to_requests(op.failure)
                request = rpc.TxnRequest(
                    compare=compare,
                    success=success_ops,
                    failure=failure_ops,
                )
                request_op = rpc.RequestOp(request_txn=request)
                request_ops.append(request_op)

            else:
                raise Exception(
                    'Unknown request class {}'.format(op.__class__))
        return request_ops

    @staticmethod
    def _build_alarm_request(alarm_action, member_id, alarm_type):
        alarm_request = rpc.AlarmRequest()

        if alarm_action == 'get':
            alarm_request.action = rpc.AlarmRequest.GET
        elif alarm_action == 'activate':
            alarm_request.action = rpc.AlarmRequest.ACTIVATE
        elif alarm_action == 'deactivate':
            alarm_request.action = rpc.AlarmRequest.DEACTIVATE
        else:
            raise ValueError('Unknown alarm action: {}'.format(alarm_action))

        alarm_request.memberID = member_id

        if alarm_type == 'none':
            alarm_request.alarm = rpc.NONE
        elif alarm_type == 'no space':
            alarm_request.alarm = rpc.NOSPACE
        else:
            raise ValueError('Unknown alarm type: {}'.format(alarm_type))

        return alarm_request


def client(
        host='localhost',
        port=2379,
        ca_cert=None,
        cert_key=None,
        cert_cert=None,
        timeout=None,
        user=None,
        password=None,
        **kwargs,
):
    """Return an instance of an Etcd3Client."""
    return Etcd3Client(
        host=host,
        port=port,
        ca_cert=ca_cert,
        cert_key=cert_key,
        cert_cert=cert_cert,
        timeout=timeout,
        user=user,
        password=password,
        **kwargs,
    )



from . import rpc
from . import utils


_OPERATORS = {
    rpc.Compare.EQUAL: '==',
    rpc.Compare.NOT_EQUAL: '!=',
    rpc.Compare.LESS: '<',
    rpc.Compare.GREATER: '>',
}


class BaseCompare(object):
    def __init__(self, key, range_end=None):
        self.key = key
        self.range_end = range_end
        self.value = None
        self.op = None

    # TODO check other is of correct type for compare
    # Version, Mod and Create can only be ints
    def __eq__(self, other):
        self.value = other
        self.op = rpc.Compare.EQUAL
        return self

    def __ne__(self, other):
        self.value = other
        self.op = rpc.Compare.NOT_EQUAL
        return self

    def __lt__(self, other):
        self.value = other
        self.op = rpc.Compare.LESS
        return self

    def __gt__(self, other):
        self.value = other
        self.op = rpc.Compare.GREATER
        return self

    def __repr__(self):
        if self.range_end is None:
            keys = self.key
        else:
            keys = f'[{self.key}, {self.range_end})'
        return f"{self.__class__}: {keys} {_OPERATORS.get(self.op)} '{self.value}'"

    def build_message(self):
        compare = rpc.Compare()
        compare.key = utils.to_bytes(self.key)
        if self.range_end is not None:
            compare.range_end = utils.to_bytes(self.range_end)

        if self.op is None:
            raise ValueError('op must be one of =, !=, < or >')

        compare.result = self.op

        self.build_compare(compare)
        return compare


class Value(BaseCompare):
    def build_compare(self, compare):
        compare.target = rpc.Compare.VALUE
        compare.value = utils.to_bytes(self.value)


class Version(BaseCompare):
    def build_compare(self, compare):
        compare.target = rpc.Compare.VERSION
        compare.version = int(self.value)


class Create(BaseCompare):
    def build_compare(self, compare):
        compare.target = rpc.Compare.CREATE
        compare.create_revision = int(self.value)


class Mod(BaseCompare):
    def build_compare(self, compare):
        compare.target = rpc.Compare.MOD
        compare.mod_revision = int(self.value)


class Put(object):
    def __init__(self, key, value, lease=None, prev_kv=False):
        self.key = key
        self.value = value
        self.lease = lease
        self.prev_kv = prev_kv


class Get(object):
    def __init__(self, key, range_end=None):
        self.key = key
        self.range_end = range_end


class Delete(object):
    def __init__(self, key, range_end=None, prev_kv=False):
        self.key = key
        self.range_end = range_end
        self.prev_kv = prev_kv


class Txn(object):
    def __init__(self, compare, success=None, failure=None):
        self.compare = compare
        self.success = success
        self.failure = failure



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: auth.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='auth.proto',
  package='authpb',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\nauth.proto\x12\x06\x61uthpb\"%\n\x0eUserAddOptions\x12\x13\n\x0bno_password\x18\x01 \x01(\x08\"^\n\x04User\x12\x0c\n\x04name\x18\x01 \x01(\x0c\x12\x10\n\x08password\x18\x02 \x01(\x0c\x12\r\n\x05roles\x18\x03 \x03(\t\x12\'\n\x07options\x18\x04 \x01(\x0b\x32\x16.authpb.UserAddOptions\"\x83\x01\n\nPermission\x12)\n\x08permType\x18\x01 \x01(\x0e\x32\x17.authpb.Permission.Type\x12\x0b\n\x03key\x18\x02 \x01(\x0c\x12\x11\n\trange_end\x18\x03 \x01(\x0c\"*\n\x04Type\x12\x08\n\x04READ\x10\x00\x12\t\n\x05WRITE\x10\x01\x12\r\n\tREADWRITE\x10\x02\"?\n\x04Role\x12\x0c\n\x04name\x18\x01 \x01(\x0c\x12)\n\rkeyPermission\x18\x02 \x03(\x0b\x32\x12.authpb.Permissionb\x06proto3'
)



_PERMISSION_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='authpb.Permission.Type',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='READ', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WRITE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='READWRITE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=247,
  serialized_end=289,
)
_sym_db.RegisterEnumDescriptor(_PERMISSION_TYPE)


_USERADDOPTIONS = _descriptor.Descriptor(
  name='UserAddOptions',
  full_name='authpb.UserAddOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='no_password', full_name='authpb.UserAddOptions.no_password', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22,
  serialized_end=59,
)


_USER = _descriptor.Descriptor(
  name='User',
  full_name='authpb.User',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='authpb.User.name', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='password', full_name='authpb.User.password', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='roles', full_name='authpb.User.roles', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='options', full_name='authpb.User.options', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=61,
  serialized_end=155,
)


_PERMISSION = _descriptor.Descriptor(
  name='Permission',
  full_name='authpb.Permission',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='permType', full_name='authpb.Permission.permType', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='key', full_name='authpb.Permission.key', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='range_end', full_name='authpb.Permission.range_end', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _PERMISSION_TYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=158,
  serialized_end=289,
)


_ROLE = _descriptor.Descriptor(
  name='Role',
  full_name='authpb.Role',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='authpb.Role.name', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='keyPermission', full_name='authpb.Role.keyPermission', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=291,
  serialized_end=354,
)

_USER.fields_by_name['options'].message_type = _USERADDOPTIONS
_PERMISSION.fields_by_name['permType'].enum_type = _PERMISSION_TYPE
_PERMISSION_TYPE.containing_type = _PERMISSION
_ROLE.fields_by_name['keyPermission'].message_type = _PERMISSION
DESCRIPTOR.message_types_by_name['UserAddOptions'] = _USERADDOPTIONS
DESCRIPTOR.message_types_by_name['User'] = _USER
DESCRIPTOR.message_types_by_name['Permission'] = _PERMISSION
DESCRIPTOR.message_types_by_name['Role'] = _ROLE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

UserAddOptions = _reflection.GeneratedProtocolMessageType('UserAddOptions', (_message.Message,), {
  'DESCRIPTOR' : _USERADDOPTIONS,
  '__module__' : 'auth_pb2'
  # @@protoc_insertion_point(class_scope:authpb.UserAddOptions)
  })
_sym_db.RegisterMessage(UserAddOptions)

User = _reflection.GeneratedProtocolMessageType('User', (_message.Message,), {
  'DESCRIPTOR' : _USER,
  '__module__' : 'auth_pb2'
  # @@protoc_insertion_point(class_scope:authpb.User)
  })
_sym_db.RegisterMessage(User)

Permission = _reflection.GeneratedProtocolMessageType('Permission', (_message.Message,), {
  'DESCRIPTOR' : _PERMISSION,
  '__module__' : 'auth_pb2'
  # @@protoc_insertion_point(class_scope:authpb.Permission)
  })
_sym_db.RegisterMessage(Permission)

Role = _reflection.GeneratedProtocolMessageType('Role', (_message.Message,), {
  'DESCRIPTOR' : _ROLE,
  '__module__' : 'auth_pb2'
  # @@protoc_insertion_point(class_scope:authpb.Role)
  })
_sym_db.RegisterMessage(Role)


# @@protoc_insertion_point(module_scope)



# Generated by the Protocol Buffers compiler. DO NOT EDIT!
# source: rpc.proto
# plugin: grpclib.plugin.main
import abc
import typing

import grpclib.const
import grpclib.client
if typing.TYPE_CHECKING:
    import grpclib.server

from . import auth_pb2
from . import kv_pb2
from . import rpc_pb2


class KVBase(abc.ABC):

    @abc.abstractmethod
    async def Range(self, stream: 'grpclib.server.Stream[rpc_pb2.RangeRequest, rpc_pb2.RangeResponse]') -> None:
        pass

    @abc.abstractmethod
    async def Put(self, stream: 'grpclib.server.Stream[rpc_pb2.PutRequest, rpc_pb2.PutResponse]') -> None:
        pass

    @abc.abstractmethod
    async def DeleteRange(self, stream: 'grpclib.server.Stream[rpc_pb2.DeleteRangeRequest, rpc_pb2.DeleteRangeResponse]') -> None:
        pass

    @abc.abstractmethod
    async def Txn(self, stream: 'grpclib.server.Stream[rpc_pb2.TxnRequest, rpc_pb2.TxnResponse]') -> None:
        pass

    @abc.abstractmethod
    async def Compact(self, stream: 'grpclib.server.Stream[rpc_pb2.CompactionRequest, rpc_pb2.CompactionResponse]') -> None:
        pass

    def __mapping__(self) -> typing.Dict[str, grpclib.const.Handler]:
        return {
            '/etcdserverpb.KV/Range': grpclib.const.Handler(
                self.Range,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.RangeRequest,
                rpc_pb2.RangeResponse,
            ),
            '/etcdserverpb.KV/Put': grpclib.const.Handler(
                self.Put,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.PutRequest,
                rpc_pb2.PutResponse,
            ),
            '/etcdserverpb.KV/DeleteRange': grpclib.const.Handler(
                self.DeleteRange,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.DeleteRangeRequest,
                rpc_pb2.DeleteRangeResponse,
            ),
            '/etcdserverpb.KV/Txn': grpclib.const.Handler(
                self.Txn,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.TxnRequest,
                rpc_pb2.TxnResponse,
            ),
            '/etcdserverpb.KV/Compact': grpclib.const.Handler(
                self.Compact,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.CompactionRequest,
                rpc_pb2.CompactionResponse,
            ),
        }


class KVStub:

    def __init__(self, channel: grpclib.client.Channel) -> None:
        self.Range = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.KV/Range',
            rpc_pb2.RangeRequest,
            rpc_pb2.RangeResponse,
        )
        self.Put = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.KV/Put',
            rpc_pb2.PutRequest,
            rpc_pb2.PutResponse,
        )
        self.DeleteRange = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.KV/DeleteRange',
            rpc_pb2.DeleteRangeRequest,
            rpc_pb2.DeleteRangeResponse,
        )
        self.Txn = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.KV/Txn',
            rpc_pb2.TxnRequest,
            rpc_pb2.TxnResponse,
        )
        self.Compact = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.KV/Compact',
            rpc_pb2.CompactionRequest,
            rpc_pb2.CompactionResponse,
        )


class WatchBase(abc.ABC):

    @abc.abstractmethod
    async def Watch(self, stream: 'grpclib.server.Stream[rpc_pb2.WatchRequest, rpc_pb2.WatchResponse]') -> None:
        pass

    def __mapping__(self) -> typing.Dict[str, grpclib.const.Handler]:
        return {
            '/etcdserverpb.Watch/Watch': grpclib.const.Handler(
                self.Watch,
                grpclib.const.Cardinality.STREAM_STREAM,
                rpc_pb2.WatchRequest,
                rpc_pb2.WatchResponse,
            ),
        }


class WatchStub:

    def __init__(self, channel: grpclib.client.Channel) -> None:
        self.Watch = grpclib.client.StreamStreamMethod(
            channel,
            '/etcdserverpb.Watch/Watch',
            rpc_pb2.WatchRequest,
            rpc_pb2.WatchResponse,
        )


class LeaseBase(abc.ABC):

    @abc.abstractmethod
    async def LeaseGrant(self, stream: 'grpclib.server.Stream[rpc_pb2.LeaseGrantRequest, rpc_pb2.LeaseGrantResponse]') -> None:
        pass

    @abc.abstractmethod
    async def LeaseRevoke(self, stream: 'grpclib.server.Stream[rpc_pb2.LeaseRevokeRequest, rpc_pb2.LeaseRevokeResponse]') -> None:
        pass

    @abc.abstractmethod
    async def LeaseKeepAlive(self, stream: 'grpclib.server.Stream[rpc_pb2.LeaseKeepAliveRequest, rpc_pb2.LeaseKeepAliveResponse]') -> None:
        pass

    @abc.abstractmethod
    async def LeaseTimeToLive(self, stream: 'grpclib.server.Stream[rpc_pb2.LeaseTimeToLiveRequest, rpc_pb2.LeaseTimeToLiveResponse]') -> None:
        pass

    @abc.abstractmethod
    async def LeaseLeases(self, stream: 'grpclib.server.Stream[rpc_pb2.LeaseLeasesRequest, rpc_pb2.LeaseLeasesResponse]') -> None:
        pass

    def __mapping__(self) -> typing.Dict[str, grpclib.const.Handler]:
        return {
            '/etcdserverpb.Lease/LeaseGrant': grpclib.const.Handler(
                self.LeaseGrant,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.LeaseGrantRequest,
                rpc_pb2.LeaseGrantResponse,
            ),
            '/etcdserverpb.Lease/LeaseRevoke': grpclib.const.Handler(
                self.LeaseRevoke,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.LeaseRevokeRequest,
                rpc_pb2.LeaseRevokeResponse,
            ),
            '/etcdserverpb.Lease/LeaseKeepAlive': grpclib.const.Handler(
                self.LeaseKeepAlive,
                grpclib.const.Cardinality.STREAM_STREAM,
                rpc_pb2.LeaseKeepAliveRequest,
                rpc_pb2.LeaseKeepAliveResponse,
            ),
            '/etcdserverpb.Lease/LeaseTimeToLive': grpclib.const.Handler(
                self.LeaseTimeToLive,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.LeaseTimeToLiveRequest,
                rpc_pb2.LeaseTimeToLiveResponse,
            ),
            '/etcdserverpb.Lease/LeaseLeases': grpclib.const.Handler(
                self.LeaseLeases,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.LeaseLeasesRequest,
                rpc_pb2.LeaseLeasesResponse,
            ),
        }


class LeaseStub:

    def __init__(self, channel: grpclib.client.Channel) -> None:
        self.LeaseGrant = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Lease/LeaseGrant',
            rpc_pb2.LeaseGrantRequest,
            rpc_pb2.LeaseGrantResponse,
        )
        self.LeaseRevoke = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Lease/LeaseRevoke',
            rpc_pb2.LeaseRevokeRequest,
            rpc_pb2.LeaseRevokeResponse,
        )
        self.LeaseKeepAlive = grpclib.client.StreamStreamMethod(
            channel,
            '/etcdserverpb.Lease/LeaseKeepAlive',
            rpc_pb2.LeaseKeepAliveRequest,
            rpc_pb2.LeaseKeepAliveResponse,
        )
        self.LeaseTimeToLive = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Lease/LeaseTimeToLive',
            rpc_pb2.LeaseTimeToLiveRequest,
            rpc_pb2.LeaseTimeToLiveResponse,
        )
        self.LeaseLeases = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Lease/LeaseLeases',
            rpc_pb2.LeaseLeasesRequest,
            rpc_pb2.LeaseLeasesResponse,
        )


class ClusterBase(abc.ABC):

    @abc.abstractmethod
    async def MemberAdd(self, stream: 'grpclib.server.Stream[rpc_pb2.MemberAddRequest, rpc_pb2.MemberAddResponse]') -> None:
        pass

    @abc.abstractmethod
    async def MemberRemove(self, stream: 'grpclib.server.Stream[rpc_pb2.MemberRemoveRequest, rpc_pb2.MemberRemoveResponse]') -> None:
        pass

    @abc.abstractmethod
    async def MemberUpdate(self, stream: 'grpclib.server.Stream[rpc_pb2.MemberUpdateRequest, rpc_pb2.MemberUpdateResponse]') -> None:
        pass

    @abc.abstractmethod
    async def MemberList(self, stream: 'grpclib.server.Stream[rpc_pb2.MemberListRequest, rpc_pb2.MemberListResponse]') -> None:
        pass

    @abc.abstractmethod
    async def MemberPromote(self, stream: 'grpclib.server.Stream[rpc_pb2.MemberPromoteRequest, rpc_pb2.MemberPromoteResponse]') -> None:
        pass

    def __mapping__(self) -> typing.Dict[str, grpclib.const.Handler]:
        return {
            '/etcdserverpb.Cluster/MemberAdd': grpclib.const.Handler(
                self.MemberAdd,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.MemberAddRequest,
                rpc_pb2.MemberAddResponse,
            ),
            '/etcdserverpb.Cluster/MemberRemove': grpclib.const.Handler(
                self.MemberRemove,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.MemberRemoveRequest,
                rpc_pb2.MemberRemoveResponse,
            ),
            '/etcdserverpb.Cluster/MemberUpdate': grpclib.const.Handler(
                self.MemberUpdate,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.MemberUpdateRequest,
                rpc_pb2.MemberUpdateResponse,
            ),
            '/etcdserverpb.Cluster/MemberList': grpclib.const.Handler(
                self.MemberList,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.MemberListRequest,
                rpc_pb2.MemberListResponse,
            ),
            '/etcdserverpb.Cluster/MemberPromote': grpclib.const.Handler(
                self.MemberPromote,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.MemberPromoteRequest,
                rpc_pb2.MemberPromoteResponse,
            ),
        }


class ClusterStub:

    def __init__(self, channel: grpclib.client.Channel) -> None:
        self.MemberAdd = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Cluster/MemberAdd',
            rpc_pb2.MemberAddRequest,
            rpc_pb2.MemberAddResponse,
        )
        self.MemberRemove = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Cluster/MemberRemove',
            rpc_pb2.MemberRemoveRequest,
            rpc_pb2.MemberRemoveResponse,
        )
        self.MemberUpdate = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Cluster/MemberUpdate',
            rpc_pb2.MemberUpdateRequest,
            rpc_pb2.MemberUpdateResponse,
        )
        self.MemberList = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Cluster/MemberList',
            rpc_pb2.MemberListRequest,
            rpc_pb2.MemberListResponse,
        )
        self.MemberPromote = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Cluster/MemberPromote',
            rpc_pb2.MemberPromoteRequest,
            rpc_pb2.MemberPromoteResponse,
        )


class MaintenanceBase(abc.ABC):

    @abc.abstractmethod
    async def Alarm(self, stream: 'grpclib.server.Stream[rpc_pb2.AlarmRequest, rpc_pb2.AlarmResponse]') -> None:
        pass

    @abc.abstractmethod
    async def Status(self, stream: 'grpclib.server.Stream[rpc_pb2.StatusRequest, rpc_pb2.StatusResponse]') -> None:
        pass

    @abc.abstractmethod
    async def Defragment(self, stream: 'grpclib.server.Stream[rpc_pb2.DefragmentRequest, rpc_pb2.DefragmentResponse]') -> None:
        pass

    @abc.abstractmethod
    async def Hash(self, stream: 'grpclib.server.Stream[rpc_pb2.HashRequest, rpc_pb2.HashResponse]') -> None:
        pass

    @abc.abstractmethod
    async def HashKV(self, stream: 'grpclib.server.Stream[rpc_pb2.HashKVRequest, rpc_pb2.HashKVResponse]') -> None:
        pass

    @abc.abstractmethod
    async def Snapshot(self, stream: 'grpclib.server.Stream[rpc_pb2.SnapshotRequest, rpc_pb2.SnapshotResponse]') -> None:
        pass

    @abc.abstractmethod
    async def MoveLeader(self, stream: 'grpclib.server.Stream[rpc_pb2.MoveLeaderRequest, rpc_pb2.MoveLeaderResponse]') -> None:
        pass

    def __mapping__(self) -> typing.Dict[str, grpclib.const.Handler]:
        return {
            '/etcdserverpb.Maintenance/Alarm': grpclib.const.Handler(
                self.Alarm,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.AlarmRequest,
                rpc_pb2.AlarmResponse,
            ),
            '/etcdserverpb.Maintenance/Status': grpclib.const.Handler(
                self.Status,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.StatusRequest,
                rpc_pb2.StatusResponse,
            ),
            '/etcdserverpb.Maintenance/Defragment': grpclib.const.Handler(
                self.Defragment,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.DefragmentRequest,
                rpc_pb2.DefragmentResponse,
            ),
            '/etcdserverpb.Maintenance/Hash': grpclib.const.Handler(
                self.Hash,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.HashRequest,
                rpc_pb2.HashResponse,
            ),
            '/etcdserverpb.Maintenance/HashKV': grpclib.const.Handler(
                self.HashKV,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.HashKVRequest,
                rpc_pb2.HashKVResponse,
            ),
            '/etcdserverpb.Maintenance/Snapshot': grpclib.const.Handler(
                self.Snapshot,
                grpclib.const.Cardinality.UNARY_STREAM,
                rpc_pb2.SnapshotRequest,
                rpc_pb2.SnapshotResponse,
            ),
            '/etcdserverpb.Maintenance/MoveLeader': grpclib.const.Handler(
                self.MoveLeader,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.MoveLeaderRequest,
                rpc_pb2.MoveLeaderResponse,
            ),
        }


class MaintenanceStub:

    def __init__(self, channel: grpclib.client.Channel) -> None:
        self.Alarm = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Maintenance/Alarm',
            rpc_pb2.AlarmRequest,
            rpc_pb2.AlarmResponse,
        )
        self.Status = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Maintenance/Status',
            rpc_pb2.StatusRequest,
            rpc_pb2.StatusResponse,
        )
        self.Defragment = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Maintenance/Defragment',
            rpc_pb2.DefragmentRequest,
            rpc_pb2.DefragmentResponse,
        )
        self.Hash = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Maintenance/Hash',
            rpc_pb2.HashRequest,
            rpc_pb2.HashResponse,
        )
        self.HashKV = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Maintenance/HashKV',
            rpc_pb2.HashKVRequest,
            rpc_pb2.HashKVResponse,
        )
        self.Snapshot = grpclib.client.UnaryStreamMethod(
            channel,
            '/etcdserverpb.Maintenance/Snapshot',
            rpc_pb2.SnapshotRequest,
            rpc_pb2.SnapshotResponse,
        )
        self.MoveLeader = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Maintenance/MoveLeader',
            rpc_pb2.MoveLeaderRequest,
            rpc_pb2.MoveLeaderResponse,
        )


class AuthBase(abc.ABC):

    @abc.abstractmethod
    async def AuthEnable(self, stream: 'grpclib.server.Stream[rpc_pb2.AuthEnableRequest, rpc_pb2.AuthEnableResponse]') -> None:
        pass

    @abc.abstractmethod
    async def AuthDisable(self, stream: 'grpclib.server.Stream[rpc_pb2.AuthDisableRequest, rpc_pb2.AuthDisableResponse]') -> None:
        pass

    @abc.abstractmethod
    async def Authenticate(self, stream: 'grpclib.server.Stream[rpc_pb2.AuthenticateRequest, rpc_pb2.AuthenticateResponse]') -> None:
        pass

    @abc.abstractmethod
    async def UserAdd(self, stream: 'grpclib.server.Stream[rpc_pb2.AuthUserAddRequest, rpc_pb2.AuthUserAddResponse]') -> None:
        pass

    @abc.abstractmethod
    async def UserGet(self, stream: 'grpclib.server.Stream[rpc_pb2.AuthUserGetRequest, rpc_pb2.AuthUserGetResponse]') -> None:
        pass

    @abc.abstractmethod
    async def UserList(self, stream: 'grpclib.server.Stream[rpc_pb2.AuthUserListRequest, rpc_pb2.AuthUserListResponse]') -> None:
        pass

    @abc.abstractmethod
    async def UserDelete(self, stream: 'grpclib.server.Stream[rpc_pb2.AuthUserDeleteRequest, rpc_pb2.AuthUserDeleteResponse]') -> None:
        pass

    @abc.abstractmethod
    async def UserChangePassword(self, stream: 'grpclib.server.Stream[rpc_pb2.AuthUserChangePasswordRequest, rpc_pb2.AuthUserChangePasswordResponse]') -> None:
        pass

    @abc.abstractmethod
    async def UserGrantRole(self, stream: 'grpclib.server.Stream[rpc_pb2.AuthUserGrantRoleRequest, rpc_pb2.AuthUserGrantRoleResponse]') -> None:
        pass

    @abc.abstractmethod
    async def UserRevokeRole(self, stream: 'grpclib.server.Stream[rpc_pb2.AuthUserRevokeRoleRequest, rpc_pb2.AuthUserRevokeRoleResponse]') -> None:
        pass

    @abc.abstractmethod
    async def RoleAdd(self, stream: 'grpclib.server.Stream[rpc_pb2.AuthRoleAddRequest, rpc_pb2.AuthRoleAddResponse]') -> None:
        pass

    @abc.abstractmethod
    async def RoleGet(self, stream: 'grpclib.server.Stream[rpc_pb2.AuthRoleGetRequest, rpc_pb2.AuthRoleGetResponse]') -> None:
        pass

    @abc.abstractmethod
    async def RoleList(self, stream: 'grpclib.server.Stream[rpc_pb2.AuthRoleListRequest, rpc_pb2.AuthRoleListResponse]') -> None:
        pass

    @abc.abstractmethod
    async def RoleDelete(self, stream: 'grpclib.server.Stream[rpc_pb2.AuthRoleDeleteRequest, rpc_pb2.AuthRoleDeleteResponse]') -> None:
        pass

    @abc.abstractmethod
    async def RoleGrantPermission(self, stream: 'grpclib.server.Stream[rpc_pb2.AuthRoleGrantPermissionRequest, rpc_pb2.AuthRoleGrantPermissionResponse]') -> None:
        pass

    @abc.abstractmethod
    async def RoleRevokePermission(self, stream: 'grpclib.server.Stream[rpc_pb2.AuthRoleRevokePermissionRequest, rpc_pb2.AuthRoleRevokePermissionResponse]') -> None:
        pass

    def __mapping__(self) -> typing.Dict[str, grpclib.const.Handler]:
        return {
            '/etcdserverpb.Auth/AuthEnable': grpclib.const.Handler(
                self.AuthEnable,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.AuthEnableRequest,
                rpc_pb2.AuthEnableResponse,
            ),
            '/etcdserverpb.Auth/AuthDisable': grpclib.const.Handler(
                self.AuthDisable,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.AuthDisableRequest,
                rpc_pb2.AuthDisableResponse,
            ),
            '/etcdserverpb.Auth/Authenticate': grpclib.const.Handler(
                self.Authenticate,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.AuthenticateRequest,
                rpc_pb2.AuthenticateResponse,
            ),
            '/etcdserverpb.Auth/UserAdd': grpclib.const.Handler(
                self.UserAdd,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.AuthUserAddRequest,
                rpc_pb2.AuthUserAddResponse,
            ),
            '/etcdserverpb.Auth/UserGet': grpclib.const.Handler(
                self.UserGet,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.AuthUserGetRequest,
                rpc_pb2.AuthUserGetResponse,
            ),
            '/etcdserverpb.Auth/UserList': grpclib.const.Handler(
                self.UserList,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.AuthUserListRequest,
                rpc_pb2.AuthUserListResponse,
            ),
            '/etcdserverpb.Auth/UserDelete': grpclib.const.Handler(
                self.UserDelete,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.AuthUserDeleteRequest,
                rpc_pb2.AuthUserDeleteResponse,
            ),
            '/etcdserverpb.Auth/UserChangePassword': grpclib.const.Handler(
                self.UserChangePassword,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.AuthUserChangePasswordRequest,
                rpc_pb2.AuthUserChangePasswordResponse,
            ),
            '/etcdserverpb.Auth/UserGrantRole': grpclib.const.Handler(
                self.UserGrantRole,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.AuthUserGrantRoleRequest,
                rpc_pb2.AuthUserGrantRoleResponse,
            ),
            '/etcdserverpb.Auth/UserRevokeRole': grpclib.const.Handler(
                self.UserRevokeRole,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.AuthUserRevokeRoleRequest,
                rpc_pb2.AuthUserRevokeRoleResponse,
            ),
            '/etcdserverpb.Auth/RoleAdd': grpclib.const.Handler(
                self.RoleAdd,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.AuthRoleAddRequest,
                rpc_pb2.AuthRoleAddResponse,
            ),
            '/etcdserverpb.Auth/RoleGet': grpclib.const.Handler(
                self.RoleGet,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.AuthRoleGetRequest,
                rpc_pb2.AuthRoleGetResponse,
            ),
            '/etcdserverpb.Auth/RoleList': grpclib.const.Handler(
                self.RoleList,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.AuthRoleListRequest,
                rpc_pb2.AuthRoleListResponse,
            ),
            '/etcdserverpb.Auth/RoleDelete': grpclib.const.Handler(
                self.RoleDelete,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.AuthRoleDeleteRequest,
                rpc_pb2.AuthRoleDeleteResponse,
            ),
            '/etcdserverpb.Auth/RoleGrantPermission': grpclib.const.Handler(
                self.RoleGrantPermission,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.AuthRoleGrantPermissionRequest,
                rpc_pb2.AuthRoleGrantPermissionResponse,
            ),
            '/etcdserverpb.Auth/RoleRevokePermission': grpclib.const.Handler(
                self.RoleRevokePermission,
                grpclib.const.Cardinality.UNARY_UNARY,
                rpc_pb2.AuthRoleRevokePermissionRequest,
                rpc_pb2.AuthRoleRevokePermissionResponse,
            ),
        }


class AuthStub:

    def __init__(self, channel: grpclib.client.Channel) -> None:
        self.AuthEnable = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Auth/AuthEnable',
            rpc_pb2.AuthEnableRequest,
            rpc_pb2.AuthEnableResponse,
        )
        self.AuthDisable = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Auth/AuthDisable',
            rpc_pb2.AuthDisableRequest,
            rpc_pb2.AuthDisableResponse,
        )
        self.Authenticate = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Auth/Authenticate',
            rpc_pb2.AuthenticateRequest,
            rpc_pb2.AuthenticateResponse,
        )
        self.UserAdd = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Auth/UserAdd',
            rpc_pb2.AuthUserAddRequest,
            rpc_pb2.AuthUserAddResponse,
        )
        self.UserGet = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Auth/UserGet',
            rpc_pb2.AuthUserGetRequest,
            rpc_pb2.AuthUserGetResponse,
        )
        self.UserList = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Auth/UserList',
            rpc_pb2.AuthUserListRequest,
            rpc_pb2.AuthUserListResponse,
        )
        self.UserDelete = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Auth/UserDelete',
            rpc_pb2.AuthUserDeleteRequest,
            rpc_pb2.AuthUserDeleteResponse,
        )
        self.UserChangePassword = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Auth/UserChangePassword',
            rpc_pb2.AuthUserChangePasswordRequest,
            rpc_pb2.AuthUserChangePasswordResponse,
        )
        self.UserGrantRole = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Auth/UserGrantRole',
            rpc_pb2.AuthUserGrantRoleRequest,
            rpc_pb2.AuthUserGrantRoleResponse,
        )
        self.UserRevokeRole = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Auth/UserRevokeRole',
            rpc_pb2.AuthUserRevokeRoleRequest,
            rpc_pb2.AuthUserRevokeRoleResponse,
        )
        self.RoleAdd = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Auth/RoleAdd',
            rpc_pb2.AuthRoleAddRequest,
            rpc_pb2.AuthRoleAddResponse,
        )
        self.RoleGet = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Auth/RoleGet',
            rpc_pb2.AuthRoleGetRequest,
            rpc_pb2.AuthRoleGetResponse,
        )
        self.RoleList = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Auth/RoleList',
            rpc_pb2.AuthRoleListRequest,
            rpc_pb2.AuthRoleListResponse,
        )
        self.RoleDelete = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Auth/RoleDelete',
            rpc_pb2.AuthRoleDeleteRequest,
            rpc_pb2.AuthRoleDeleteResponse,
        )
        self.RoleGrantPermission = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Auth/RoleGrantPermission',
            rpc_pb2.AuthRoleGrantPermissionRequest,
            rpc_pb2.AuthRoleGrantPermissionResponse,
        )
        self.RoleRevokePermission = grpclib.client.UnaryUnaryMethod(
            channel,
            '/etcdserverpb.Auth/RoleRevokePermission',
            rpc_pb2.AuthRoleRevokePermissionRequest,
            rpc_pb2.AuthRoleRevokePermissionResponse,
        )



class Member:
    """A member of the etcd cluster.

    :param id: ID of the cluster member
    :param name: Human-readable name of the cluster member
    :param peer_urls: List of URLs the cluster member exposes to the cluster for
                     communication
    :param client_urls: List of URLs the cluster member exposes to clients for
                       communication
    :param etcd_client: Instance of :class:`aetcd3.client.Etcd3Client`
    """

    def __init__(self, id, name, peer_urls, client_urls, etcd_client):
        self.id = id
        self.name = name
        self.peer_urls = peer_urls
        self.client_urls = client_urls

        self._etcd_client = etcd_client

    def __str__(self):
        return (
            f'Cluster member with ID: {self.id}, peer URLs: {self.peer_urls}, client '
            f'URLs: {self.client_urls}'
        )

    async def remove(self):
        """Remove this cluster member from the cluster."""
        await self._etcd_client.remove_member(self.id)

    async def update(self, peer_urls):
        """Update the configuration of this cluster member.

        :param peer_urls: New list of peer URLs the cluster member will use to
                          communicate with the cluster
        """
        await self._etcd_client.update_member(self.id, peer_urls)

    async def active_alarms(self):
        """Get active alarms of the cluster member.

        :returns: List of :class:`aetcd3.client.Alarm`
        """
        return await self._etcd_client.list_alarms(member_id=self.id)



from . import rpc
from .client import Alarm
from .client import Etcd3Client
from .client import Transactions
from .client import client
from .events import DeleteEvent
from .events import Event
from .events import PutEvent
from .exceptions import Etcd3Exception
from .leases import Lease
from .locks import Lock
from .members import Member


__all__ = (
    'Alarm',
    'DeleteEvent',
    'Etcd3Client',
    'Etcd3Exception',
    'Event',
    'Lease',
    'Lock',
    'Member',
    'PutEvent',
    'Transactions',
    'client',
    'rpc',
)



class Etcd3Exception(Exception):
    """Raises on generic client errors."""


class WatchTimedOut(Etcd3Exception):
    """Raised on operation timeouts."""


class InternalServerError(Etcd3Exception):
    """Raises on etcd internal server errors."""


class ConnectionFailedError(Etcd3Exception):
    """Raises on etcd server connection errors."""

    def __str__(self):
        return 'etcd connection failed'


class ConnectionTimeoutError(Etcd3Exception):
    """Raises on etcd server connection timeout errors."""

    def __str__(self):
        return 'etcd connection timeout'


class PreconditionFailedError(Etcd3Exception):
    """Raises on etcd server precondition errors."""


class RevisionCompactedError(Etcd3Exception):
    """Raises when requested and previous revisions were already compacted.

    :param compacted_revision: Revision bellow values were compacted
    :type compacted_revision: int
    """

    def __init__(self, compacted_revision):
        self.compacted_revision = compacted_revision
        super(RevisionCompactedError, self).__init__()



class Lease(object):
    """A lease.

    :param id: ID of the lease
    :param ttl: time to live for this lease
    :param etcd_client: Instance of :class:`aetcd3.Etcd3Client`
    """

    def __init__(self, lease_id, ttl, etcd_client):
        self.id = lease_id
        self.ttl = ttl

        self._etcd_client = etcd_client

    async def _get_lease_info(self, *, keys=True):
        return await self._etcd_client.get_lease_info(self.id, keys=keys)

    async def revoke(self):
        """Revoke this lease."""
        await self._etcd_client.revoke_lease(self.id)

    async def refresh(self):
        """Refresh the time to live for this lease."""
        return await self._etcd_client.refresh_lease(self.id)

    async def remaining_ttl(self):
        """Return remaining time to live for this lease."""
        return (await self._get_lease_info(keys=False)).TTL

    async def granted_ttl(self):
        """Return granted time to live for this lease."""
        return (await self._get_lease_info(keys=False)).grantedTTL

    async def keys(self):
        """Return list of keys associated with this lease."""
        return (await self._get_lease_info()).keys



class Event(object):
    """Base event type.

    :param event: Raw gRPC event
    """

    def __init__(self, event):
        self.key = event.kv.key

        self._event = event

    def __getattr__(self, name):
        if name.startswith('prev_'):
            return getattr(self._event.prev_kv, name[5:])
        return getattr(self._event.kv, name)

    def __str__(self):
        return f'{self.__class__} key={self.key} value={self.value}'


class PutEvent(Event):
    """Put event type."""


class DeleteEvent(Event):
    """Delete event type."""


def new_event(event):
    """Wrap a raw gRPC event in a friendlier containing class.

    This picks the appropriate class from one of :class:`PutEvent` or
    :class:`DeleteEvent` and returns a new instance.

    If wrong raw gRPC event was provided `Exception` is raised.

    :param event: Raw gRPC event

    :rtype: One of :class:`PutEvent` or :class:`DeleteEvent`

    :raises: :class:`Exception`
    """
    op_name = event.EventType.DESCRIPTOR.values_by_number[event.type].name
    if op_name == 'PUT':
        cls = PutEvent
    elif op_name == 'DELETE':
        cls = DeleteEvent
    else:
        raise Exception('Invalid gRPC event type name')

    return cls(event)



import asyncio
import base64
import contextlib
import json
import os
import signal
import string
import subprocess
import tempfile
import threading
import time
import unittest.mock
import urllib.parse

import grpclib
import pytest
import tenacity

import aetcd3.exceptions
import aetcd3.rpc as rpc
import aetcd3.utils as utils


etcd_version = os.environ.get('TEST_ETCD_VERSION', 'v3.2.8')

os.environ['ETCDCTL_API'] = '3'


def etcdctl(*args):
    endpoint = os.environ.get('PYTHON_ETCD_HTTP_URL')
    if endpoint:
        args = ['--endpoints', endpoint] + list(args)
    args = ['etcdctl', '-w', 'json'] + list(args)
    print(' '.join(args))
    output = subprocess.check_output(args)
    return json.loads(output.decode('utf-8'))


@contextlib.contextmanager
def _out_quorum():
    pids = subprocess.check_output(['pgrep', '-f', '--', '--name pifpaf[12]'])
    pids = [int(pid.strip()) for pid in pids.splitlines()]
    try:
        for pid in pids:
            os.kill(pid, signal.SIGSTOP)
        yield
    finally:
        for pid in pids:
            os.kill(pid, signal.SIGCONT)


class TestEtcd3:
    class MockedException(grpclib.exceptions.GRPCError):
        def __init__(self, status):
            self.status = status

    @pytest.fixture
    async def etcd(self):
        endpoint = os.environ.get('PYTHON_ETCD_HTTP_URL')
        timeout = 5
        if endpoint:
            url = urllib.parse.urlparse(endpoint)
            with aetcd3.client(
                host=url.hostname,
                port=url.port,
                timeout=timeout,
            ) as client:
                yield client
        else:
            async with aetcd3.client() as client:
                yield client

        @tenacity.retry(wait=tenacity.wait_fixed(2), stop=tenacity.stop_after_attempt(3))
        def delete_keys_definitely():
            # clean up after fixture goes out of scope
            etcdctl('del', '--prefix', '/')
            out = etcdctl('get', '--prefix', '/')
            assert 'kvs' not in out

        delete_keys_definitely()

    @pytest.mark.asyncio
    async def test_get_unknown_key(self, etcd):
        value, meta = await etcd.get('probably-invalid-key')
        assert value is None
        assert meta is None

    @pytest.mark.asyncio
    async def test_get_key(self, etcd, string='xxx'):
        etcdctl('put', '/doot/a_key', string)
        returned, _ = await etcd.get('/doot/a_key')
        assert returned == string.encode('utf-8')

    @pytest.mark.asyncio
    async def test_get_random_key(self, etcd, string='xxxx'):
        etcdctl('put', '/doot/' + string, 'dootdoot')
        returned, _ = await etcd.get('/doot/' + string)
        assert returned == b'dootdoot'

    @pytest.mark.asyncio
    async def test_get_have_cluster_revision(self, etcd, string='xxx'):
        etcdctl('put', '/doot/' + string, 'dootdoot')
        _, md = await etcd.get('/doot/' + string)
        assert md.response_header.revision > 0

    # @given(characters(blacklist_categories=['Cs', 'Cc']))
    @pytest.mark.asyncio
    async def test_put_key(self, etcd, string='xxx'):
        await etcd.put('/doot/put_1', string)
        out = etcdctl('get', '/doot/put_1')
        assert base64.b64decode(out['kvs'][0]['value']) == \
               string.encode('utf-8')

    # @given(
    #     characters(blacklist_categories=['Cs', 'Cc']),
    #     characters(blacklist_categories=['Cs', 'Cc']),
    # )
    @pytest.mark.asyncio
    async def test_get_key_serializable(self, etcd, key='foo', string='xxx'):
        etcdctl('put', '/doot/' + key, string)
        with _out_quorum():
            returned, _ = await etcd.get('/doot/' + key, serializable=True)
        assert returned == string.encode('utf-8')

    # @given(characters(blacklist_categories=['Cs', 'Cc']))
    @pytest.mark.asyncio
    async def test_put_has_cluster_revision(self, etcd, string='xxx'):
        response = await etcd.put('/doot/put_1', string)
        assert response.header.revision > 0

    # @given(characters(blacklist_categories=['Cs', 'Cc']))
    @pytest.mark.asyncio
    async def test_put_has_prev_kv(self, etcd, string='xxxx'):
        etcdctl('put', '/doot/put_1', 'old_value')
        response = await etcd.put('/doot/put_1', string, prev_kv=True)
        assert response.prev_kv.value == b'old_value'

    @pytest.mark.asyncio
    async def test_delete_key(self, etcd):
        etcdctl('put', '/doot/delete_this', 'delete pls')

        v, _ = await etcd.get('/doot/delete_this')
        assert v == b'delete pls'

        deleted = await etcd.delete('/doot/delete_this')
        assert deleted is True

        deleted = await etcd.delete('/doot/delete_this')
        assert deleted is False

        deleted = await etcd.delete('/doot/not_here_dude')
        assert deleted is False

        v, _ = await etcd.get('/doot/delete_this')
        assert v is None

    @pytest.mark.asyncio
    async def test_delete_has_cluster_revision(self, etcd):
        response = await etcd.delete('/doot/delete_this', return_response=True)
        assert response.header.revision > 0

    @pytest.mark.asyncio
    async def test_delete_has_prev_kv(self, etcd):
        etcdctl('put', '/doot/delete_this', 'old_value')
        response = await etcd.delete('/doot/delete_this',
                                     prev_kv=True,
                                     return_response=True)
        assert response.prev_kvs[0].value == b'old_value'

    @pytest.mark.asyncio
    async def test_delete_keys_with_prefix(self, etcd):
        etcdctl('put', '/foo/1', 'bar')
        etcdctl('put', '/foo/2', 'baz')

        v, _ = await etcd.get('/foo/1')
        assert v == b'bar'

        v, _ = await etcd.get('/foo/2')
        assert v == b'baz'

        response = await etcd.delete_prefix('/foo')
        assert response.deleted == 2

        v, _ = await etcd.get('/foo/1')
        assert v is None

        v, _ = await etcd.get('/foo/2')
        assert v is None

    @pytest.mark.asyncio
    async def test_watch_key(self, etcd):
        def update_etcd(v):
            etcdctl('put', '/doot/watch', v)
            out = etcdctl('get', '/doot/watch')
            assert base64.b64decode(out['kvs'][0]['value']) == \
                   utils.to_bytes(v)

        def update_key():
            # sleep to make watch can get the event
            time.sleep(3)
            update_etcd('0')
            time.sleep(1)
            update_etcd('1')
            time.sleep(1)
            update_etcd('2')
            time.sleep(1)
            update_etcd('3')
            time.sleep(1)

        t = threading.Thread(name='update_key', target=update_key)
        t.start()

        change_count = 0
        events_iterator, cancel = await etcd.watch(b'/doot/watch')
        async for event in events_iterator:
            assert event.key == b'/doot/watch'
            assert event.value == utils.to_bytes(str(change_count))

            # if cancel worked, we should not receive event 3
            assert event.value != utils.to_bytes('3')

            change_count += 1
            if change_count > 2:
                # if cancel not work, we will block in this for-loop forever
                await cancel()

        t.join()

    @pytest.mark.asyncio
    async def test_watch_key_with_revision_compacted(self, etcd):
        etcdctl('put', '/watchcompation', '0')  # Some data to compact
        value, meta = await etcd.get('/watchcompation')
        revision = meta.mod_revision

        # Compact etcd and test watcher
        await etcd.compact(revision)

        def update_etcd(v):
            etcdctl('put', '/watchcompation', v)
            out = etcdctl('get', '/watchcompation')
            assert base64.b64decode(out['kvs'][0]['value']) == \
                   utils.to_bytes(v)

        def update_key():
            update_etcd('1')
            update_etcd('2')
            update_etcd('3')

        t = threading.Thread(name='update_key', target=update_key)
        t.start()

        async def watch_compacted_revision_test():
            events_iterator, cancel = await etcd.watch(
                b'/watchcompation', start_revision=(revision - 1))

            error_raised = False
            compacted_revision = 0
            try:
                await events_iterator.__anext__()
            except Exception as err:
                error_raised = True
                assert isinstance(err,
                                  aetcd3.exceptions.RevisionCompactedError)
                compacted_revision = err.compacted_revision

            assert error_raised is True
            assert compacted_revision == revision

            change_count = 0
            events_iterator, cancel = await etcd.watch(
                b'/watchcompation', start_revision=compacted_revision)
            async for event in events_iterator:
                assert event.key == b'/watchcompation'
                assert event.value == \
                       utils.to_bytes(str(change_count))

                # if cancel worked, we should not receive event 3
                assert event.value != utils.to_bytes('3')

                change_count += 1
                if change_count > 2:
                    await cancel()

        await watch_compacted_revision_test()

        t.join()

    @pytest.mark.asyncio
    async def test_watch_exception_during_watch(self, etcd):
        await etcd.open()

        async def pass_exception_to_callback(callback):
            await asyncio.sleep(1)
            ex = self.MockedException(grpclib.const.Status.UNAVAILABLE)
            await callback(ex)

        task = None

        async def add_callback_mock(*args, **kwargs):
            nonlocal task
            callback = args[1]
            task = asyncio.get_event_loop().create_task(
                pass_exception_to_callback(callback))
            return 1

        watcher_mock = unittest.mock.MagicMock()
        watcher_mock.add_callback = add_callback_mock
        etcd.watcher = watcher_mock

        events_iterator, cancel = await etcd.watch('foo')

        with pytest.raises(aetcd3.exceptions.ConnectionFailedError):
            async for _ in events_iterator:
                _

        await task

    @pytest.mark.asyncio
    async def test_watch_timeout_on_establishment(self):
        async with aetcd3.client(timeout=3) as foo_etcd:
            @contextlib.asynccontextmanager
            async def slow_watch_mock(*args, **kwargs):
                await asyncio.sleep(40)
                yield 'foo'

            foo_etcd.watcher._watch_stub.Watch.open = slow_watch_mock  # noqa

            with pytest.raises(aetcd3.exceptions.WatchTimedOut):
                events_iterator, cancel = await foo_etcd.watch('foo')
                async for _ in events_iterator:
                    pass

    @pytest.mark.asyncio
    async def test_watch_prefix(self, etcd):
        def update_etcd(v):
            etcdctl('put', '/doot/watch/prefix/' + v, v)
            out = etcdctl('get', '/doot/watch/prefix/' + v)
            assert base64.b64decode(out['kvs'][0]['value']) == \
                   utils.to_bytes(v)

        def update_key():
            # sleep to make watch can get the event
            time.sleep(3)
            update_etcd('0')
            time.sleep(1)
            update_etcd('1')
            time.sleep(1)
            update_etcd('2')
            time.sleep(1)
            update_etcd('3')
            time.sleep(1)

        t = threading.Thread(name='update_key_prefix', target=update_key)
        t.start()

        change_count = 0
        events_iterator, cancel = await etcd.watch_prefix(
            '/doot/watch/prefix/')
        async for event in events_iterator:
            assert event.key == \
                   utils.to_bytes('/doot/watch/prefix/{}'.format(change_count))
            assert event.value == \
                   utils.to_bytes(str(change_count))

            # if cancel worked, we should not receive event 3
            assert event.value != utils.to_bytes('3')

            change_count += 1
            if change_count > 2:
                # if cancel not work, we will block in this for-loop forever
                await cancel()

        t.join()

    @pytest.mark.asyncio
    async def test_sequential_watch_prefix_once(self, etcd):
        try:
            await etcd.watch_prefix_once('/doot/', 1)
        except aetcd3.exceptions.WatchTimedOut:
            print('timeout1')
            pass
        try:
            await etcd.watch_prefix_once('/doot/', 1)
        except aetcd3.exceptions.WatchTimedOut:
            print('timeout2')
            pass
        try:
            await etcd.watch_prefix_once('/doot/', 1)
        except aetcd3.exceptions.WatchTimedOut:
            print('timeout3')
            pass

    @pytest.mark.asyncio
    async def test_transaction_success(self, etcd):
        etcdctl('put', '/doot/txn', 'dootdoot')
        await etcd.transaction(
            compare=[etcd.transactions.value('/doot/txn') == 'dootdoot'],
            success=[etcd.transactions.put('/doot/txn', 'success')],
            failure=[etcd.transactions.put('/doot/txn', 'failure')],
        )
        out = etcdctl('get', '/doot/txn')
        assert base64.b64decode(out['kvs'][0]['value']) == b'success'

    @pytest.mark.asyncio
    async def test_transaction_failure(self, etcd):
        etcdctl('put', '/doot/txn', 'notdootdoot')
        await etcd.transaction(
            compare=[etcd.transactions.value('/doot/txn') == 'dootdoot'],
            success=[etcd.transactions.put('/doot/txn', 'success')],
            failure=[etcd.transactions.put('/doot/txn', 'failure')],
        )
        out = etcdctl('get', '/doot/txn')
        assert base64.b64decode(out['kvs'][0]['value']) == b'failure'

    def test_ops_to_requests(self, etcd):
        with pytest.raises(Exception):
            etcd._ops_to_requests(['not_transaction_type'])
        with pytest.raises(TypeError):
            etcd._ops_to_requests(0)

    @pytest.mark.skipif(etcd_version < 'v3.3',
                        reason='requires etcd v3.3 or higher')
    @pytest.mark.asyncio
    async def test_nested_transactions(self, etcd):
        await etcd.transaction(
            compare=[],
            success=[etcd.transactions.put('/doot/txn1', '1'),
                     etcd.transactions.txn(
                         compare=[],
                         success=[etcd.transactions.put('/doot/txn2', '2')],
                         failure=[])],
            failure=[],
        )
        value, _ = await etcd.get('/doot/txn1')
        assert value == b'1'
        value, _ = await etcd.get('/doot/txn2')
        assert value == b'2'

    @pytest.mark.asyncio
    async def test_replace_success(self, etcd):
        await etcd.put('/doot/thing', 'toot')
        status = await etcd.replace('/doot/thing', 'toot', 'doot')
        v, _ = await etcd.get('/doot/thing')
        assert v == b'doot'
        assert status is True

    @pytest.mark.asyncio
    async def test_replace_fail(self, etcd):
        await etcd.put('/doot/thing', 'boot')
        status = await etcd.replace('/doot/thing', 'toot', 'doot')
        v, _ = await etcd.get('/doot/thing')
        assert v == b'boot'
        assert status is False

    @pytest.mark.asyncio
    async def test_get_prefix(self, etcd):
        for i in range(20):
            etcdctl('put', '/doot/range{}'.format(i), 'i am a range')

        for i in range(5):
            etcdctl('put', '/doot/notrange{}'.format(i), 'i am a not range')

        values = [p async for p in etcd.get_prefix('/doot/range')]
        assert len(values) == 20
        for value, _ in values:
            assert value == b'i am a range'

    @pytest.mark.asyncio
    async def test_get_prefix_keys_only(self, etcd):
        for i in range(20):
            etcdctl('put', '/doot/range{}'.format(i), 'i am a range')

        for i in range(5):
            etcdctl('put', '/doot/notrange{}'.format(i), 'i am a not range')

        values = [p async for p in etcd.get_prefix('/doot/range',
                                                   keys_only=True)]
        assert len(values) == 20
        for value, meta in values:
            assert meta.key.startswith(b'/doot/range')
            assert not value

    @pytest.mark.asyncio
    async def test_get_range(self, etcd):
        for char in string.ascii_lowercase:
            if char < 'p':
                etcdctl('put', '/doot/' + char, 'i am in range')
            else:
                etcdctl('put', '/doot/' + char, 'i am not in range')

        values = [v async for v in etcd.get_range('/doot/a', '/doot/p')]
        assert len(values) == 15
        for value, _ in values:
            assert value == b'i am in range'

    @pytest.mark.asyncio
    async def test_all_not_found_error(self, etcd):
        result = [x async for x in etcd.get_all()]
        assert not result

    @pytest.mark.asyncio
    async def test_range_not_found_error(self, etcd):
        for i in range(5):
            etcdctl('put', '/doot/notrange{}'.format(i), 'i am a not range')

        result = [p async for p in etcd.get_prefix('/doot/range')]
        assert not result

    @pytest.mark.asyncio
    async def test_get_all(self, etcd):
        for i in range(20):
            etcdctl('put', '/doot/range{}'.format(i), 'i am in all')

        for i in range(5):
            etcdctl('put', '/doot/notrange{}'.format(i), 'i am in all')
        values = [x async for x in etcd.get_all()]
        assert len(values) == 25
        for value, _ in values:
            assert value == b'i am in all'

    @pytest.mark.asyncio
    async def test_sort_order(self, etcd):
        def remove_prefix(string, prefix):
            return string[len(prefix):]

        initial_keys = 'abcde'
        initial_values = 'qwert'

        for k, v in zip(initial_keys, initial_values):
            etcdctl('put', '/doot/{}'.format(k), v)

        keys = ''
        async for value, meta in etcd.get_prefix('/doot', sort_order='ascend'):
            keys += remove_prefix(meta.key.decode('utf-8'), '/doot/')

        assert keys == initial_keys

        reverse_keys = ''
        async for value, meta in etcd.get_prefix('/doot',
                                                 sort_order='descend'):
            reverse_keys += remove_prefix(meta.key.decode('utf-8'), '/doot/')

        assert reverse_keys == ''.join(reversed(initial_keys))

    @pytest.mark.asyncio
    async def test_lease_grant(self, etcd):
        lease = await etcd.lease(1)

        assert isinstance(lease.ttl, int)
        assert isinstance(lease.id, int)

    @pytest.mark.asyncio
    async def test_lease_revoke(self, etcd):
        lease = await etcd.lease(1)
        await lease.revoke()

    @pytest.mark.skipif(etcd_version.startswith('v3.0'),
                        reason='requires etcd v3.1 or higher')
    @pytest.mark.asyncio
    async def test_lease_keys_empty(self, etcd):
        lease = await etcd.lease(1)
        assert (await lease.keys()) == []

    @pytest.mark.skipif(etcd_version.startswith('v3.0'),
                        reason='requires etcd v3.1 or higher')
    @pytest.mark.asyncio
    async def test_lease_single_key(self, etcd):
        lease = await etcd.lease(1)
        await etcd.put('/doot/lease_test', 'this is a lease', lease=lease)
        assert (await lease.keys()) == [b'/doot/lease_test']

    @pytest.mark.skipif(etcd_version.startswith('v3.0'),
                        reason='requires etcd v3.1 or higher')
    @pytest.mark.asyncio
    async def test_lease_expire(self, etcd):
        key = '/doot/lease_test_expire'
        lease = await etcd.lease(1)
        await etcd.put(key, 'this is a lease', lease=lease)
        assert (await lease.keys()) == [utils.to_bytes(key)]
        v, _ = await etcd.get(key)
        assert v == b'this is a lease'
        assert (await lease.remaining_ttl()) <= (await lease.granted_ttl())

        # wait for the lease to expire
        await asyncio.sleep((await lease.granted_ttl()) + 2)
        v, _ = await etcd.get(key)
        assert v is None

    @pytest.mark.asyncio
    async def test_member_list(self, etcd):
        assert len([m async for m in etcd.members()]) == 3
        async for member in etcd.members():
            assert member.name.startswith('pifpaf')
            for peer_url in member.peer_urls:
                assert peer_url.startswith('http://')
            for client_url in member.client_urls:
                assert client_url.startswith('http://')
            assert isinstance(member.id, int) is True

    @pytest.mark.asyncio
    async def test_lock_acquire(self, etcd):
        lock = etcd.lock('lock-1', ttl=10)
        assert (await lock.acquire()) is True
        assert (await etcd.get(lock.key))[0] is not None
        assert (await lock.acquire(timeout=0)) is False
        assert (await lock.acquire(timeout=1)) is False

    @pytest.mark.asyncio
    async def test_lock_release(self, etcd):
        lock = etcd.lock('lock-2', ttl=10)
        assert (await lock.acquire()) is True
        assert (await etcd.get(lock.key))[0] is not None
        assert (await lock.release()) is True
        v, _ = await etcd.get(lock.key)
        assert v is None
        assert (await lock.acquire()) is True
        assert (await lock.release()) is True
        assert (await lock.acquire(timeout=None)) is True

    @pytest.mark.asyncio
    async def test_lock_expire(self, etcd):
        lock = etcd.lock('lock-3', ttl=3)
        assert (await lock.acquire()) is True
        assert (await etcd.get(lock.key))[0] is not None
        # wait for the lease to expire
        await asyncio.sleep(9)
        v, _ = await etcd.get(lock.key)
        assert v is None

    @pytest.mark.asyncio
    async def test_lock_refresh(self, etcd):
        lock = etcd.lock('lock-4', ttl=3)
        assert (await lock.acquire()) is True
        assert (await etcd.get(lock.key))[0] is not None
        # sleep for the same total time as test_lock_expire, but refresh each
        # second
        for _ in range(9):
            await asyncio.sleep(1)
            await lock.refresh()

        assert (await etcd.get(lock.key))[0] is not None

    @pytest.mark.asyncio
    async def test_lock_is_acquired(self, etcd):
        lock1 = etcd.lock('lock-5', ttl=2)
        assert (await lock1.is_acquired()) is False

        lock2 = etcd.lock('lock-5', ttl=2)
        await lock2.acquire()
        assert (await lock2.is_acquired()) is True
        await lock2.release()

        lock3 = etcd.lock('lock-5', ttl=2)
        await lock3.acquire()
        assert (await lock3.is_acquired()) is True
        assert (await lock2.is_acquired()) is False

    @pytest.mark.asyncio
    async def test_lock_context_manager(self, etcd):
        async with etcd.lock('lock-6', ttl=2) as lock:
            assert (await lock.is_acquired()) is True
        assert (await lock.is_acquired()) is False

    @pytest.mark.asyncio
    async def test_lock_contended(self, etcd):
        lock1 = etcd.lock('lock-7', ttl=2)
        await lock1.acquire()
        lock2 = etcd.lock('lock-7', ttl=2)
        await lock2.acquire()
        assert (await lock1.is_acquired()) is False
        assert (await lock2.is_acquired()) is True

    @pytest.mark.asyncio
    async def test_lock_double_acquire_release(self, etcd):
        lock = etcd.lock('lock-8', ttl=10)
        assert (await lock.acquire(0)) is True
        assert (await lock.acquire(0)) is False
        assert (await lock.release()) is True

    @pytest.mark.asyncio
    async def test_lock_acquire_none(self, etcd):
        lock = etcd.lock('lock-9', ttl=10)
        assert (await lock.acquire(None)) is True
        # This will succeed after 10 seconds since the TTL will expire and the
        # lock is not refreshed
        assert (await lock.acquire(None)) is True

    @pytest.mark.asyncio
    async def test_internal_exception_on_internal_error(self, etcd):
        await etcd.open()
        exception = self.MockedException(grpclib.const.Status.INTERNAL)
        kv_mock = unittest.mock.MagicMock()
        kv_mock.Range.side_effect = exception
        etcd.kvstub = kv_mock

        with pytest.raises(aetcd3.exceptions.InternalServerError):
            await etcd.get('foo')

    @pytest.mark.asyncio
    async def test_connection_failure_exception_on_connection_failure(self, etcd):
        await etcd.open()
        exception = self.MockedException(grpclib.const.Status.UNAVAILABLE)
        kv_mock = unittest.mock.MagicMock()
        kv_mock.Range.side_effect = exception
        etcd.kvstub = kv_mock

        with pytest.raises(aetcd3.exceptions.ConnectionFailedError):
            await etcd.get('foo')

    @pytest.mark.asyncio
    async def test_connection_timeout_exception_on_connection_timeout(self, etcd):
        ex = self.MockedException(grpclib.const.Status.DEADLINE_EXCEEDED)

        class MockKvstub:
            async def Range(self, *args, **kwargs):  # noqa: N802
                raise ex

        etcd.kvstub = MockKvstub()

        with pytest.raises(aetcd3.exceptions.ConnectionTimeoutError):
            await etcd.get('foo')

    @pytest.mark.asyncio
    async def test_grpc_exception_on_unknown_code(self, etcd):
        exception = self.MockedException(grpclib.const.Status.DATA_LOSS)
        kv_mock = unittest.mock.MagicMock()
        kv_mock.Range.side_effect = exception
        etcd.kvstub = kv_mock

        try:
            await etcd.get('foo')
        except grpclib.exceptions.GRPCError:
            pass
        else:
            raise RuntimeError

    @pytest.mark.asyncio
    async def test_status_member(self, etcd):
        status = await etcd.status()

        assert isinstance(status.leader, aetcd3.members.Member) is True
        assert status.leader.id in [m.id async for m in etcd.members()]

    @pytest.mark.asyncio
    async def test_hash(self, etcd):
        assert isinstance((await etcd.hash()), int)

    @pytest.mark.asyncio
    async def test_snapshot(self, etcd):
        with tempfile.NamedTemporaryFile() as f:
            await etcd.snapshot(f)
            f.flush()

            etcdctl('snapshot', 'status', f.name)


class TestAlarms(object):
    @pytest.fixture
    async def etcd(self):
        etcd = aetcd3.client()
        yield etcd
        await etcd.disarm_alarm()
        async for m in etcd.members():
            if m.active_alarms:
                await etcd.disarm_alarm(m.id)

    @pytest.mark.asyncio
    async def test_create_alarm_all_members(self, etcd):
        alarms = await etcd.create_alarm()

        assert len(alarms) == 1
        assert alarms[0].member_id == 0
        assert alarms[0].alarm_type == rpc.NOSPACE

    @pytest.mark.asyncio
    async def test_create_alarm_specific_member(self, etcd):
        members = [m async for m in etcd.members()]
        a_member = members[0]

        alarms = await etcd.create_alarm(member_id=a_member.id)

        assert len(alarms) == 1
        assert alarms[0].member_id == a_member.id
        assert alarms[0].alarm_type == rpc.NOSPACE

    @pytest.mark.asyncio
    async def test_list_alarms(self, etcd):
        members = [m async for m in etcd.members()]
        a_member = members[0]
        await etcd.create_alarm()
        await etcd.create_alarm(member_id=a_member.id)
        possible_member_ids = [0, a_member.id]

        alarms = [a async for a in etcd.list_alarms()]

        assert len(alarms) == 2
        for alarm in alarms:
            possible_member_ids.remove(alarm.member_id)
            assert alarm.alarm_type == rpc.NOSPACE

        assert possible_member_ids == []

    @pytest.mark.asyncio
    async def test_disarm_alarm(self, etcd):
        await etcd.create_alarm()
        assert len([a async for a in etcd.list_alarms()]) == 1

        await etcd.disarm_alarm()
        assert len([a async for a in etcd.list_alarms()]) == 0


class TestUtils(object):
    def test_prefix_range_end(self):
        assert aetcd3.utils.prefix_range_end(b'foo') == b'fop'

    def test_to_bytes(self):
        assert isinstance(aetcd3.utils.to_bytes(b'doot'), bytes) is True
        assert isinstance(aetcd3.utils.to_bytes('doot'), bytes) is True
        assert aetcd3.utils.to_bytes(b'doot') == b'doot'
        assert aetcd3.utils.to_bytes('doot') == b'doot'


class TestClient(object):
    @pytest.fixture
    def etcd(self):
        yield aetcd3.client()

    def test_sort_target(self, etcd):
        key = 'key'.encode('utf-8')
        sort_target = {
            None: rpc.RangeRequest.KEY,
            'key': rpc.RangeRequest.KEY,
            'version': rpc.RangeRequest.VERSION,
            'create': rpc.RangeRequest.CREATE,
            'mod': rpc.RangeRequest.MOD,
            'value': rpc.RangeRequest.VALUE,
        }

        for input, expected in sort_target.items():
            range_request = etcd._build_get_range_request(key,
                                                          sort_target=input)
            assert range_request.sort_target == expected
        with pytest.raises(ValueError):
            etcd._build_get_range_request(key, sort_target='feelsbadman')

    def test_sort_order(self, etcd):
        key = 'key'.encode('utf-8')
        sort_target = {
            None: rpc.RangeRequest.NONE,
            'ascend': rpc.RangeRequest.ASCEND,
            'descend': rpc.RangeRequest.DESCEND,
        }

        for input, expected in sort_target.items():
            range_request = etcd._build_get_range_request(key,
                                                          sort_order=input)
            assert range_request.sort_order == expected
        with pytest.raises(ValueError):
            etcd._build_get_range_request(key, sort_order='feelsbadman')

    @pytest.mark.asyncio
    async def test_secure_channel(self):
        client = aetcd3.client(
            ca_cert='tests/ca.crt',
            cert_key='tests/client.key',
            cert_cert='tests/client.crt',
        )
        await client.open()

        assert client.uses_secure_channel is True

    @pytest.mark.asyncio
    async def test_secure_channel_ca_cert_only(self):
        with tempfile.NamedTemporaryFile() as certfile_bundle:
            for fname in ('client.crt', 'ca.crt', 'client.key'):
                with open(f'tests/{fname}', 'r+b') as f:
                    certfile_bundle.write(f.read())
            certfile_bundle.flush()
            client = aetcd3.client(
                ca_cert=certfile_bundle.name,
                cert_key=None,
                cert_cert=None,
            )
            await client.open()

            assert client.uses_secure_channel is True

    def test_secure_channel_ca_cert_and_key_raise_exception(self):
        with pytest.raises(ValueError):
            aetcd3.client(
                ca_cert='tests/ca.crt',
                cert_key='tests/client.crt',
                cert_cert=None,
            )

        with pytest.raises(ValueError):
            aetcd3.client(
                ca_cert='tests/ca.crt',
                cert_key=None,
                cert_cert='tests/client.crt',
            )

    @pytest.mark.asyncio
    async def test_compact(self, etcd):
        await etcd.put('/foo', 'x')
        _, meta = await etcd.get('/foo')
        revision = meta.mod_revision
        await etcd.compact(revision)
        with pytest.raises(grpclib.exceptions.GRPCError):
            await etcd.compact(revision)

    @pytest.mark.asyncio
    async def test_channel_with_no_cert(self):
        client = aetcd3.client(
            ca_cert=None,
            cert_key=None,
            cert_cert=None,
        )
        await client.open()

        assert client.uses_secure_channel is False

    @pytest.mark.asyncio
    async def test_user_pwd_auth(self):
        with self._enabled_auth_in_etcd():
            # Create a client using username and password auth
            client = aetcd3.client(
                user='root',
                password='pwd',
            )
            await client.get('probably-invalid-key')

    def test_user_or_pwd_auth_raises_exception(self):
        with pytest.raises(Exception, match='both user and password'):
            aetcd3.client(user='usr')

        with pytest.raises(Exception, match='both user and password'):
            aetcd3.client(password='pwd')

    @staticmethod
    @contextlib.contextmanager
    def _enabled_auth_in_etcd():
        subprocess.call(['etcdctl', '-w', 'json', 'user', 'add', 'root:pwd'])
        subprocess.call(['etcdctl', 'auth', 'enable'])
        try:
            yield
        finally:
            subprocess.call(['etcdctl',
                             '-w', 'json', '--user', 'root:pwd',
                             'auth', 'disable'])
            subprocess.call(['etcdctl', 'user', 'delete', 'root'])
            subprocess.call(['etcdctl', 'role', 'delete', 'root'])


class TestCompares(object):

    def test_compare_version(self):
        key = 'key'
        tx = aetcd3.Transactions()

        version_compare = tx.version(key) == 1
        assert version_compare.op == rpc.Compare.EQUAL

        version_compare = tx.version(key) != 2
        assert version_compare.op == rpc.Compare.NOT_EQUAL

        version_compare = tx.version(key) < 91
        assert version_compare.op == rpc.Compare.LESS

        version_compare = tx.version(key) > 92
        assert version_compare.op == rpc.Compare.GREATER
        assert version_compare.build_message().target == \
               rpc.Compare.VERSION

    def test_compare_value(self):
        key = 'key'
        tx = aetcd3.Transactions()

        value_compare = tx.value(key) == 'b'
        assert value_compare.op == rpc.Compare.EQUAL

        value_compare = tx.value(key) != 'b'
        assert value_compare.op == rpc.Compare.NOT_EQUAL

        value_compare = tx.value(key) < 'b'
        assert value_compare.op == rpc.Compare.LESS

        value_compare = tx.value(key) > 'b'
        assert value_compare.op == rpc.Compare.GREATER
        assert value_compare.build_message().target == rpc.Compare.VALUE

    def test_compare_mod(self):
        key = 'key'
        tx = aetcd3.Transactions()

        mod_compare = tx.mod(key) == -100
        assert mod_compare.op == rpc.Compare.EQUAL

        mod_compare = tx.mod(key) != -100
        assert mod_compare.op == rpc.Compare.NOT_EQUAL

        mod_compare = tx.mod(key) < 19
        assert mod_compare.op == rpc.Compare.LESS

        mod_compare = tx.mod(key) > 21
        assert mod_compare.op == rpc.Compare.GREATER
        assert mod_compare.build_message().target == rpc.Compare.MOD

    def test_compare_create(self):
        key = 'key'
        tx = aetcd3.Transactions()

        create_compare = tx.create(key) == 10
        assert create_compare.op == rpc.Compare.EQUAL

        create_compare = tx.create(key) != 10
        assert create_compare.op == rpc.Compare.NOT_EQUAL

        create_compare = tx.create(key) < 155
        assert create_compare.op == rpc.Compare.LESS

        create_compare = tx.create(key) > -12
        assert create_compare.op == rpc.Compare.GREATER
        assert create_compare.build_message().target == rpc.Compare.CREATE



import asyncio
import logging

import grpclib

from . import events
from . import exceptions
from . import rpc
from . import utils


_log = logging.getLogger(__name__)


class Watch(object):

    def __init__(self, watch_id, iterator=None, etcd_client=None):
        self.watch_id = watch_id
        self.etcd_client = etcd_client
        self.iterator = iterator

    async def cancel(self):
        await self.etcd_client.cancel_watch(self.watch_id)

    def iterator(self):
        if self.iterator is not None:
            return self.iterator

        raise ValueError('Undefined iterator')


class Watcher(object):

    def __init__(self, watchstub, timeout=None, call_credentials=None,
                 metadata=None):
        self.timeout = timeout
        self._watch_stub = watchstub
        self._credentials = call_credentials
        self._metadata = metadata

        self._lock = asyncio.Lock()
        self._request_queue = asyncio.Queue(maxsize=10)
        self._callbacks = {}
        self._sender_task = None
        self._receiver_task = None
        self._new_watch_cond = asyncio.Condition(lock=self._lock)
        self._new_watch = None

    async def add_callback(self, key, callback, range_end=None,
                           start_revision=None, progress_notify=False,
                           filters=None, prev_kv=False):
        rq = self._create_watch_request(key, range_end=range_end,
                                        start_revision=start_revision,
                                        progress_notify=progress_notify,
                                        filters=filters, prev_kv=prev_kv)

        async with self._lock:
            await self._setup_stream()

            # Only one create watch request can be pending at a time, so if
            # there one already, then wait for it to complete first.
            while self._new_watch:
                await self._new_watch_cond.wait()

            # Submit a create watch request.
            new_watch = _NewWatch(callback)
            await self._request_queue.put(rq)
            self._new_watch = new_watch

            try:
                # Wait for the request to be completed, or timeout.
                try:
                    await asyncio.wait_for(self._new_watch_cond.wait(), self.timeout)
                except asyncio.TimeoutError:
                    raise exceptions.WatchTimedOut()

                # If the request not completed yet, then raise a timeout
                # exception.
                if new_watch.id is None and new_watch.err is None:
                    raise exceptions.WatchTimedOut()

                # Raise an exception if the watch request failed.
                if new_watch.err:
                    raise new_watch.err
            finally:
                # Wake up threads stuck on add_callback call if any.
                self._new_watch = None
                self._new_watch_cond.notify_all()

            return new_watch.id

    async def cancel(self, watch_id):
        async with self._lock:
            callback = self._callbacks.pop(watch_id, None)
            if not callback:
                return

            await self._cancel_no_lock(watch_id)

    async def _setup_stream(self):
        if self._sender_task:
            return

        async def sender_task(timeout, metadata, run_receiver_fn):
            async with self._watch_stub.Watch.open(timeout=timeout,
                                                   metadata=metadata) as stream:
                while True:
                    request = await self._request_queue.get()

                    if request is None:
                        break

                    try:
                        await stream.send_message(request)
                        await run_receiver_fn(stream)
                    except grpclib.exceptions.StreamTerminatedError as err:
                        await self._handle_steam_termination(err)

                await stream.end()

        self._sender_task = asyncio.get_running_loop().create_task(
            sender_task(self.timeout, self._metadata, self._run_receiver),
        )

    async def _run_receiver(self, stream):
        if self._receiver_task:
            return

        async def receiver_task():
            nonlocal stream
            try:
                while True:
                    response = await stream.recv_message()

                    if not response:
                        break

                    await self._handle_response(response)
            except grpclib.exceptions.StreamTerminatedError as err:
                await self._handle_steam_termination(err)

        self._receiver_task = asyncio.get_running_loop().create_task(receiver_task())

    async def _handle_steam_termination(self, err):
        async with self._lock:
            if self._new_watch:
                self._new_watch.err = err
                self._new_watch_cond.notify_all()

            callbacks = self._callbacks
            self._callbacks = {}

            # Rotate request queue. This way we can terminate one gRPC
            # stream and initiate another one whilst avoiding a race
            # between them over requests in the queue.
            await self._request_queue.put(None)
            self._request_queue = asyncio.Queue(maxsize=10)
        for callback in callbacks.values():
            await _safe_callback(callback, err)

    def close(self):
        """Close the streams for sending and receiving watch events."""
        for t in (self._receiver_task, self._sender_task):
            if t:
                t.cancel()

    async def _handle_response(self, rs):
        async with self._lock:
            if rs.created:
                # If the new watch request has already expired then cancel the
                # created watch right away.
                if not self._new_watch:
                    await self._cancel_no_lock(rs.watch_id)
                    return

                if rs.compact_revision != 0:
                    self._new_watch.err = exceptions.RevisionCompactedError(
                        rs.compact_revision)
                    return

                self._callbacks[rs.watch_id] = self._new_watch.callback
                self._new_watch.id = rs.watch_id
                self._new_watch_cond.notify_all()

            callback = self._callbacks.get(rs.watch_id)

        # Ignore leftovers from canceled watches.
        if not callback:
            return

        # The watcher can be safely reused, but adding a new event
        # to indicate that the revision is already compacted
        # requires api change which would break all users of this
        # module. So, raising an exception if a watcher is still
        # alive.
        if rs.compact_revision != 0:
            err = exceptions.RevisionCompactedError(rs.compact_revision)
            await _safe_callback(callback, err)
            await self.cancel(rs.watch_id)
            return

        for event in rs.events:
            await _safe_callback(callback, events.new_event(event))

    async def _cancel_no_lock(self, watch_id):
        cancel_watch = rpc.WatchCancelRequest()
        cancel_watch.watch_id = watch_id
        rq = rpc.WatchRequest(cancel_request=cancel_watch)
        await self._request_queue.put(rq)

    @staticmethod
    def _create_watch_request(key, range_end=None, start_revision=None,
                              progress_notify=False, filters=None,
                              prev_kv=False):
        create_watch = rpc.WatchCreateRequest()
        create_watch.key = utils.to_bytes(key)
        if range_end is not None:
            create_watch.range_end = utils.to_bytes(range_end)
        if start_revision is not None:
            create_watch.start_revision = start_revision
        if progress_notify:
            create_watch.progress_notify = progress_notify
        if filters is not None:
            create_watch.filters = filters
        if prev_kv:
            create_watch.prev_kv = prev_kv
        return rpc.WatchRequest(create_request=create_watch)


class _NewWatch(object):
    def __init__(self, callback):
        self.callback = callback
        self.id = None
        self.err = None


async def _safe_callback(callback, event_or_err):
    try:
        await callback(event_or_err)

    except Exception:
        _log.exception('Watch callback failed')



import setuptools


def _get_long_description():
    with open('README.rst') as readme_file:
        return readme_file.read()


setuptools.setup(
    name='aetcd3',
    use_scm_version=True,
    description='Python asyncio-based client for etcd',
    long_description=_get_long_description(),
    long_description_content_type='text/x-rst',
    author='Andrey Martyanov',
    author_email='andrey@martyanov.com',
    url='https://github.com/martyanov/aetcd3',
    packages=[
        'aetcd3',
        'aetcd3.rpc',
    ],
    package_dir={
        'aetcd3': 'aetcd3',
        'aetcd3.rpc': 'aetcd3/rpc',
    },
    include_package_data=True,
    license='Apache Software License 2.0',
    zip_safe=False,
    keywords='etcd3',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    project_urls={
        'Documentation': 'https://aetcd3.readthedocs.io',
        'Code': 'https://github.com/martyanov/aetcd3',
        'Issues': 'https://github.com/martyanov/aetcd3/issues',
    },
    python_requires='>=3.8,<4.0',
    setup_requires=[
        'setuptools_scm==6.0.1',
    ],
    install_requires=[
        'aiofiles>=0.5,<0.7',
        'grpclib>=0.4.1,<0.5',
        'protobuf>=3,<4',
        'tenacity>=6,<7',
    ],
    extras_require={
        'dev': [
            'flake8-commas==2.0.0',
            'flake8-docstrings==1.6.0',
            'flake8-isort==4.0.0',
            'flake8-quotes==3.2.0',
            'flake8==3.9.1',
            'grpcio-tools==1.37.0',
            'pep8-naming==0.11.1',
            'twine==3.4.1',
        ],
        'doc': [
            'sphinx==3.5.4',
            'sphinx_rtd_theme==0.5.2',
        ],
        'test': [
            'pifpaf==3.1.5',
            'pytest-asyncio==0.15.1',
            'pytest-cov==2.11.1',
            'pytest==6.2.3',
        ],
    },
)
